from avnBase.avn_base import  spellCorrect, preVisnet, spaceOut,merge_emoji_tokens, cleanDetNum,cleanText,semanticEdges,edges2DF, docRetoken,MaverickCoref,convert_specific_unicode,uniSimilarNodesDF
from visBase.vis_base import visnet
import networkx as nx
import spacy
import pandas as pd 
from tqdm import tqdm
from datetime import datetime
import time
from typing import List, Optional,Dict
import sys

nlp = spacy.load("en_core_web_sm")
nlp.add_pipe("custom_sentencizer", before="parser") 



def print_progress(current, total, bar_length=20,printText ='Generating AAVN Edges'):
    percent = current / total * 100
    filled_length = int(bar_length * current / total)
    bar = '█' * filled_length + '-' * (bar_length - filled_length)

    sys.stdout.write(
        f'\r{printText}: |{bar}| {percent:6.2f}% Completed. '
    )
    sys.stdout.flush()


from spacy.tokenizer import Tokenizer
# Ensure the spaCy tokenizer doesn't split the emoji components
def custom_tokenizer(nlp):
    # This ensures that ZWJ sequences are kept together as one token
    return Tokenizer(nlp.vocab, rules=nlp.Defaults.tokenizer_exceptions)


def semanticX(data:pd.DataFrame | List[str], text_column='text',nlpModel=None, spellCheck = True,nonCheckWords:List[str] = ['Atsaniik'], coreference = False,corefPath = None,
              det=None, valid_particles =None, number_words=True, clean_emoji=False, 
              heavy_number = False,verbPhrase = False, extraClean = False,
              reADJ:List[str]=['meilide'], reADV:List[str]=['piaoliangdi'],reVERB:List[str]=['caiquxingdong'], 
              reNOUN:List[str]=['Natural Language Processing'],reENT = True,hyphenated = True,
              NEG_Lemmas:set = {"no", "not", "n't", "never", "none", "nothing", "nobody", "neither","nor"},
              nonSemantic = 0, NN = 'notInclude',
              sentiScore = True, 
              UNICODE_TO_ASCII:dict =  {'’': "'", '‘': "'", '“': '"',  '”': '"',  '–': '-', '—': '-',  '…': '...',},
              copulaBridge = True,
              verbose = True
              ):
    """turn text to AAVN semantic edges 

    Args:
        data (list or dataframe): list of text or dataframe has text column 
        text_column (str, optional): text column name. Defaults to 'text'.
        nlpModel(spacy language model pipline): Defauts to  nlp = spacy.load("en_core_web_sm")
        spellCheck(bool): corret the words spellings , Defuats to True.
        nonCheckWords (list): the words are not to be corrected, 1 word by 1 and case sensative such as Peng Yang should be ['Peng','Yang'] not ['peng','yang'] or ['Peng Yang']
        coreference (bool, optional): if coreferee text. Defaults to False.  # pip install torch transformers hydra-core pytorch-lightning, pip install --no-deps git+https://github.com/Atsaniik/maverick-coref.git
        corefPath (_type_, optional): the model path. Defaults to None.    # hf download sapienzanlp/maverick-mes-preco --local-dir ./maverick-mes-preco # hf download sapienzanlp/maverick-mes-ontonotes --local-dir ./maverick-ontonotes
        det (list, optional):det. if None, Defaults to ["a", "an", "the"].
        valid_particles(set): if None, valid_particles = {
                            "up", "off", "away", "out", "over", "back",
                            "down", "on", "in", "with", "into", "by", "after"
                                                                            }
        number_words (bool, optional): remove numbers but keep such as G-8, Covi-19. Defaults to True.
        clean_emoji (bool, optional): remove emoji. Defaults to False.
        heavy_number (bool, optional): remove all numbers. Defaults to False.
        verbPhrase (bool, optional): if true, then 'take it off' beacome 'take off it' in the new text. Defaults to False.
        extraClean (bool, optional): Remove ( xxx ) parentheses and everything inside them. Defaults to False.
        reADJ (list, optional): retokenzie specific adjective. Defaults to ['meilide']. case-sensitive 
        reADV (list, optional): retokenzie specific adverb. Defaults to ['piaoliang'].  case-sensitive 
        reVERB (list, optional): retokenzie specific verb. Defaults to ['caiquxingdong']. case-sensitive 
        reNOUN (list, optional): retokenzie specific noun. Defaults to ['Natural Language Processing']. case-sensitive 
        reENT (bool, optional): retokenzie specific enity words. Defaults to True.
        hyphenated (bool, optional): retokenzie specific hyphenated words. Defaults to True.
        NEG_Lemmas (dict, optional): negative words to detect the edge weight postive or negative. Defaults to {"no", "not", "n't", "never", "none", "nothing", "nobody", "neither","nor"}.
        nonSemantic(digit,optional): if nonSemantic = 1, then generate avn edges of non semantic text adj-adv, adj-noun, adv-verb, verb-noun 
        NN: (vaule: [inlcude, only , notInclude] ) if only then only noun-noun type nonSemantic edges, if notInclude then A-A-V, if include the all A-A-V-N edge
        sentiScore (bool, optional): add sentiment score of edge. Defaults to True.
        reNameNode (bool, optional): add A-A-V-N label to each word ends. Defaults to True.
        UNICODE_TO_ASCII (dict, optional): change the unicode to ascii of English puncs. Defaults to {'’': "'", '‘': "'", '“': '"',  '”': '"',  '–': '-', '—': '-',  '…': '...',}.
        copulaBridge (bool): Finland is beautiful add one extra edge Finland-beautiful connected by copula Be , otherwise only Finland-be, be-beautiful 
        verbose(bool,optional): print the raw text and processed one 
    Returns:
        dataframes: edgesDF,eueDF(email, url,emoji of each row or item of original data)
    """
    if nlpModel:
        nlp = nlpModel
    else:
        nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("custom_sentencizer", before="parser") 
    
    # --- 1. Normalize Input to DataFrame ---
    if isinstance(data, list):
        # Convert list to DataFrame, no metadata
        df = pd.DataFrame({text_column: data})
        metadata_cols = [] 
    elif isinstance(data, pd.DataFrame):
        df = data.copy() # Work on a copy
        if text_column not in df.columns:
            raise ValueError(f"Column '{text_column}' not found in the DataFrame.")
        # Identify metadata columns (everything except the text)
        metadata_cols = [col for col in df.columns if col != text_column]
    else:
        raise TypeError("Input 'data' must be either a list of strings or a pandas DataFrame.")

    # --- 2. Process Rows ---
    df_list = [] # To collect small DataFrames (one per row)
    
    
    emails = [] 
    urls = []
    emojis = []
    verbphrases = []
    total_rows = len(df)
    for i,(index, row) in enumerate(df.iterrows(),start=1):
        print_progress(i, total_rows)
        raw_text = row[text_column]
        if verbose:
            print(f'raw_text ({i})  {raw_text}')
        raw_text = spaceOut(raw_text)
        #print('space out text', raw_text)
        # Safety check: ensure text is a string
        if not isinstance(raw_text, str):
            continue
        raw_text = convert_specific_unicode(raw_text,UNICODE_TO_ASCII=UNICODE_TO_ASCII )
        # A. Preprocessing
        # Uses 'raw_text' from the row, NOT 'texts' from the old list

        if spellCheck:
            raw_text = spellCorrect(raw_text, nonCheckWords=nonCheckWords)
            #print(f'correct text ({i})  ',raw_text)
        else:
            pass
        if coreference:
            raw_text = MaverickCoref.resolve(raw_text,device=None, path=corefPath)
        else:
            pass
        
        #text = cleanDetNum(text, nlp)
        final_text, unique_Verbphrases, email, url, emoji = cleanText(raw_text, nlp, det=det, valid_particles =valid_particles,
                                                                         number_words=number_words, clean_emoji=clean_emoji, heavy_number = heavy_number,
                                                                         verbPhrase = verbPhrase, extraClean = extraClean,
                                                                         reADJ=reADJ, reADV=reADV,
                                                                         reVERB=reVERB, reNOUN=reNOUN,
                                                                         reENT = reENT,hyphenated = hyphenated)
        #print('final_text',final_text)
        if verbose:
            print(f'processed_text ({i})  {final_text}')
        emails.append(email)
        urls.append(url)
        emojis.append(emoji)
        verbphrases.append(unique_Verbphrases)
        # B. Tokenization
        
        doc = docRetoken(final_text, nlp,reADJ=reADJ, reADV=reADV,
                         reVERB=reVERB + unique_Verbphrases, reNOUN=reNOUN,
                         reENT = reENT,hyphenated = hyphenated)
        #print('doc_text',doc.text)
        doc = merge_emoji_tokens(doc)
        # C. Edge Generation
        current_row_edges = []
        for sent in doc.sents:
            #print('sent.text',sent.text)
            #sent = nlp(sent.text)
            # semanticEdges returns a SET of strings like {'word-word (POS)', ...}
            edges_set = semanticEdges(sent, NEG_Lemmas = NEG_Lemmas,
                            nonSemantic = nonSemantic, NN = NN,copulaBridge=copulaBridge)
            #print('edges_set',edges_set)
            current_row_edges.extend(list(edges_set))
        
        # Only proceed if edges were found in this row
        if current_row_edges:
            # D. Convert strings to a DataFrame using your existing helper
            temp_df = edges2DF(current_row_edges,sentiScore = sentiScore)
            
            # E. Add Metadata (if any exist)
            for meta_col in metadata_cols:
                temp_df[meta_col] = row[meta_col]
            
            df_list.append(temp_df)

    # --- 3. Combine Results ---
    if df_list:
        edgesDF = pd.concat(df_list, ignore_index=True)
        eueDF = pd.DataFrame({'emials':emails,'urls':urls,'emojis':emojis,'verbphrases':verbphrases})
    else:
        # Return empty structures if no edges found
        return pd.DataFrame(),pd.DataFrame()
    return edgesDF,eueDF




def nounKPT(G,max_topic):
    """ topic generated by the high degree of noun and its edges 
    Args:
        G (_type_): _description_
        max_topic (int, optional): _description_. Defaults to 5.

    Returns:
        _type_: _description_
    """
    # Node Degrees
    nodeDegree = dict(G.degree(weight=True))
    nodeDegreeDF = pd.DataFrame({'node': list(nodeDegree.keys()), 'nodeDegree': list(nodeDegree.values())}) 
    nodeDegreeDF = nodeDegreeDF.sort_values(by='nodeDegree', key=lambda x: x.abs(), ascending=False)

    # Extract Topics
    central_nouns = []
    associate_words = []
    noun_edges = []
    i = 0
    
    for node in nodeDegreeDF.node:
        if isinstance(node,str):
            if '_2noun' in node:
                i += 1
                if i <max_topic:
                    central_nouns.append(node)
                    associate_words.append(G.adj[node])
                    #print('G.adj[node]',G.adj[node])
                    noun_edges0 = []
                    for neighbor, attr in G.adj[node].items():
                        weight = attr['weight']
                        noun_edges0.append((node,neighbor,weight))
                        #print("node:", neighbor, "weight:", weight)
                    noun_edges.append(noun_edges0)

    topicDF = pd.DataFrame({'central noun': central_nouns, 'associate_words': associate_words,'noun_edges':noun_edges})
    return topicDF


import networkx as nx
import pandas as pd

def communKPT(G, min_weight=1, k_clique=3, louvain_resolution = 1, method='louvain', topicN=True, maxNOUNtopic=5):
    """
    Generates communities using either K-clique percolation or Louvain detection.
    Extracts edges preserving the original source-target order from G.
    
    Parameters:
    - method: 'k-clique' or 'louvain'
    """
    # 1. Filter edges based on weight
    strong_edges = [
        (u, v) for u, v, d in G.edges(data=True) 
        if abs(d.get('weight', 1)) >= min_weight
    ]
     
    G2 = G.to_undirected()
    # Create the filtered graph for community detection
    G_filtered = G2.edge_subgraph(strong_edges).copy()
    
    # 2. Find Communities based on selected method
    try:
        if method == 'louvain':
            # Louvain algorithm (resolves to partitions of the graph)
            communities = nx.community.louvain_communities(G_filtered, weight='weight',resolution = louvain_resolution)
        else:
            # Original K-Clique algorithm
            communities = nx.community.k_clique_communities(G_filtered, k=k_clique)
        
        communities = list(communities)
        source_G = G_filtered
        
    except Exception as e:
        # Fallback to the original graph if filtering causes issues or k-clique fails
        if method == 'louvain':
            communities = list(nx.community.louvain_communities(G, weight='weight',resolution= louvain_resolution))
        else:
            communities = list(nx.community.k_clique_communities(G, k=k_clique))
        source_G = G

    # Sort communities by size (largest first)
    communities = sorted(communities, key=len, reverse=True)

    # 3. Handle Topic Generation
    if topicN:
        # Assuming nounKPT is defined elsewhere in your workspace
        topicDFnoun = nounKPT(source_G, max_topic=maxNOUNtopic)
    else:
        topicDFnoun = pd.DataFrame()

    community_edges = []
    community_is = []
    community_nodes = []

    # 4. Extract Data preserving Edge Orientation
    for i, nodes in enumerate(communities):
        nodes_set = set(nodes)
        edges_with_data = []
        
        # Iterate over ORIGINAL G edges to maintain 'starting -> targeting' order
        for u, v, data in G.edges(data=True):
            if u in nodes_set and v in nodes_set:
                weight = data.get('weight', None)
                if weight is not None:
                    edges_with_data.append((u, v, weight))
                else:
                    edges_with_data.append((u, v))
                    
        community_edges.append(edges_with_data)
        community_is.append(i)
        community_nodes.append(list(nodes))
        
    communityDF = pd.DataFrame({
        'community_no': community_is, 
        'community_nodes': community_nodes,
        'community_edges': community_edges
    })
    
    return communityDF, topicDFnoun


emoji_pattern = r'[\U00010000-\U0010ffff]'

def has_emoji(series):
    return series.str.contains(emoji_pattern, na=False, regex=True)



def netKPT(edgesDF:pd.DataFrame,col1 = 'node1',col2 = 'node2',col1a ='node1a',col2a='node2a',
            colorNOUN = 'skyblue', shapeNOUN= 'box',
            colorADJ = 'gold', shapeADJ = 'triangle',
            colorVERB = 'pink', shapeVERB = 'square',
            colorADV ='brown', shapeADV = 'star',
            zeroWeightC = 'black',posWeightC ='green',negWeightC ='red',
            topicBYnoun = True, maxNOUNtopic = 5,
            commMethod='louvain' ,louvain_resolution=1,
            minWeightTopic=1,kCliqueCOMMtopic = 2, 
            lowerText = True, nodeMinLen =3, keep_emoji = True,
            removeNodes:List[str] = ['when','how','why','what', 'who', 'around', 'also','that','this','these','those','you','mine','ours','their','her','his','round','go','be','have','is','are','was','were','get'],
            unifySimliarNodes = True, similarPOS ='NOUN',minSimilar = 0.99, manual_simPair:Dict[str, str] = {'suomi_2noun': 'finland_2noun'},
            visualizeNET = False,
            visNoPOS = True,
            vis_network_title = 'Atsaniik',
            vis_description_df: Optional[pd.DataFrame] = None,
            vis_description_title = "Introduce your network",
            vis_writeHTML = "network_visualization.html",
            vis_browserView= False,
            vis_min_default_node_size= -1000,
            vis_min_default_edge_width = -1000,
            vis_maximum_display = 100):
    """semantic edges to be cleaned and weighted for creating topics 

    Args:
        edgesDF (dataframe): semantic edges dataframe 
        col1 (str, 'good_2adj'): full starting node column name. Defauts to 'node1'
        col2 (str,'food_2noun'): full targeting node column name. Defauts to 'node2
        col1a (str,'good'): starting node column name. Defauts to 'node1a'
        col2a (str,'food'): targeting node column name. Defauts to 'node2a'
        colorNOUN (str, optional): _description_. Defaults to 'blue'.
        shapeNOUN (str, optional): _description_. Defaults to 'box'.
        colorADJ (str, optional): _description_. Defaults to 'gold'.
        shapeADJ (str, optional): _description_. Defaults to 'triangle'.
        colorVERB (str, optional): _description_. Defaults to 'pink'.
        shapeVERB (str, optional): _description_. Defaults to 'square'.
        colorADV (str, optional): _description_. Defaults to 'brown'.
        shapeADV (str, optional): _description_. Defaults to 'star'.
        zeroWeightC (str, optional): _description_. Defaults to 'black'.
        posWeightC (str, optional): _description_. Defaults to 'green'.
        negWeightC (str, optional): _description_. Defaults to 'red'.
        topicBYnoun(bool): if true also generate topic by Noun by centrality degree 
        maxNOUNtopic (int, optional): _description_. Defaults to 5.
        method= commMethod: community generated by 'k-clique' or 'louvain'
        louvain_resolution: louvain method resolution
        minWeightTopic (int, optional): set minimal weight absoluate value. Defaults to 1. greater than 1 or less than 1. Defaults to 1.
        kCliqueCOMMtopic (int, optional): comunnity clique number. Defaults to 3. Defaults to 3.
        lowerText (bool, clean edges): lower all nodes . Defaults to True.
        nodeMinLen (int, clean edges):  node text minimun length. Defaults to 3. 
        keep_emoji(bool,clean edges): if keep or remove the emoji edges 
        removeNodes (list, clean edges): remove edges that have the specific nodes. Defaults to ['when','how','why','what','that','this','these','those','you','mine','ours','their','her','his'].
        unifySimliarNodes ( bool ): if unify the simliar nodes in the edges, Defualts to True , only take noun and verb. 
        similarPOS (str, unify nodes by pos): Defautls to 'NOUN',
        minSimilar (float, unify nodes minimal similarity): Defauts to 0.91 # food drink 0.909 
        manual_simPair(list of dict): {'helsinki_2noun': 'finland_2noun'}
        visualizeNET (bool): if visualize the semantic network
        visNoPOS(bool): if true, then 'good_2adj' to 'good' in vis
        network_title (str): Defauts to 'Atsaniik'
        vis_description_df (str): Defausts to None
        vis_description_title (str): Defaults to "Introduce your network"
        vis_writeHTML: str = "network_visualization.html"
        vis_browserView: bool = False,
        vis_min_default_node_size: int = 0
        vis_min_default_edge_width: int = 0
        vis_maximum_display: int = 100
    Returns:
        topicDFnoun: topic by noun by centrality degree; 
        topicDFcomm: topic by community;
        edgesDF1: clean edgesDF by lowerText, nodeMinLen, and removeModes;
        edgesDF_vis: edgesDF weighted by edge groups for visualization;
        visNodes, visEdges : prepare for visnet visualization of the whole network;
    """
    
    pbar = tqdm(total=6, desc="Initializing Pipeline")
    # Combine: Remove if either column meets the criteria
    if lowerText:
        edgesDF = edgesDF.map(lambda x: x.lower() if isinstance(x, str) else x)
     
       
    if keep_emoji:

        mask1 = (
            (edgesDF[col1a].str.len() < nodeMinLen) & ~has_emoji(edgesDF[col1a])
        ) | (edgesDF[col1a].isin(removeNodes))

        mask2 = (
            (edgesDF[col2a].str.len() < nodeMinLen) & ~has_emoji(edgesDF[col2a])
        ) | (edgesDF[col2a].isin(removeNodes))
        
        edgesDF1 = edgesDF[~(mask1 | mask2)]
        
    else:
        
       mask1 = (edgesDF[col1a].str.len() < nodeMinLen) | (edgesDF[col1a].isin(removeNodes))
       mask2 = (edgesDF[col2a].str.len() < nodeMinLen) | (edgesDF[col2a].isin(removeNodes))
       edgesDF1 = edgesDF[~(mask1 | mask2)]    
       
    pbar.update(1)
    pbar.set_description("Grouping and aggregating edges")   
    
    if unifySimliarNodes:
        # col1 ='node1',col2='node2', similarPOS ='NOUN',minSimilar = 0.91
        edgesDF1,similarNodesDF = uniSimilarNodesDF(edgesDF1, col1 =col1,col2= col2, pos =similarPOS,minSimilar =minSimilar,manual_pairs=manual_simPair)
    
        edgesDF_global = edgesDF1.groupby([col1, col2]).agg(weight=('weight', 'sum')).reset_index()
    else:
        edgesDF_global = edgesDF1.groupby([col1, col2]).agg(weight=('weight', 'sum')).reset_index()
    
    
    
    pbar.update(1)
    pbar.set_description("Constructing NetworkX Graph")
    
    
    G = nx.DiGraph() 
    weighted_edge_list = list(zip(edgesDF_global[col1], edgesDF_global[col2], edgesDF_global['weight']))
    G.add_weighted_edges_from(weighted_edge_list)
    
    # Node Degrees
    nodeDegree = dict(G.degree(weight="weight")) # type: ignore
    nodeDegreeNode0 = list(nodeDegree.keys())
    nodeDegreeNode1a = [x.split('_2')[0] for x in nodeDegreeNode0]
    nodeDegreeNode1b = [x.split('_2')[1] for x in nodeDegreeNode0]
    
    nodeDegreeDF = pd.DataFrame({'node': nodeDegreeNode0, 'node1': nodeDegreeNode1a, 'nodeDegree': list(nodeDegree.values()),'pos': nodeDegreeNode1b}) 
    nodeDegreeDF = nodeDegreeDF.sort_values(by='nodeDegree', ascending=False)
    
    #for u, v, data in G.edges(data=True):
        #print('G1-u-v',u,v)
    pbar.update(1)
    pbar.set_description("Processing Node attributes")

    # --- 6. Visualization Data ---
    visNodes = []
    for node, degree in zip(nodeDegreeDF.iloc[:, 0], nodeDegreeDF.iloc[:, 2]):
        
        if "_2noun" in node:
            color = colorNOUN; shape= shapeNOUN; title= node
        elif "_2adj" in node:
            color = colorADJ; shape = shapeADJ; title = node
        elif "_2verb" in node:
            color = colorVERB; shape = shapeVERB; title = node
        elif "_2adv" in node:
            color =colorADV; shape = shapeADV; title = node 
        else:
            color = 'black'; shape = 'dot'; title = 'non AAVN'
        if visNoPOS:
            visNode = {"id": node, "label": node.split("_2")[0], "size": degree, "color": color, "shape": shape, "title": f"{title}-{degree}"}
        else:
            visNode = {"id": node, "label": node, "size": degree, "color": color, "shape": shape, "title": f"{title}-{degree}"}
        visNodes.append(visNode)
    pbar.update(1)
    pbar.set_description("Processing Edge attributes")
        
    visEdges = []
  
    # Calculate global sentiment for coloring
    global_sentiment = edgesDF1.groupby([col1, col2]).agg(sentiment=('sentiment', lambda x: x.mode().iat[0])).reset_index()
    edges_viz = pd.merge(edgesDF_global, global_sentiment, on=[col1,col2])

    for node1, node2, weight, sentiment in zip(edges_viz[col1], edges_viz[col2], edges_viz['weight'], edges_viz['sentiment']):
        if weight == 0:
            color = zeroWeightC
        elif weight > 0:
            color =posWeightC
        else:
            color = negWeightC
            
        visEdge = {"from": node1, "to": node2, "width": abs(weight), "color": {"color": color}, "arrows": "to", "title": f"weight-sentiment: {weight}-{sentiment}"}
        visEdges.append(visEdge)
    
    pbar.update(1)
    pbar.set_description("Generating topics") 
        
    #topicDFnoun = nounKPT(G,max_topic=maxNOUNtopic) 
    topicDFcomm,topicDFnoun = communKPT(G, method= commMethod,louvain_resolution=louvain_resolution,min_weight=minWeightTopic, k_clique = kCliqueCOMMtopic,topicN = topicBYnoun , maxNOUNtopic=maxNOUNtopic)
    if visualizeNET:
        sentiments2 = []
        for edge in visEdges:
            sentiments2.append (float(edge['title'].split('-')[-1]))
            
        visDF = pd.DataFrame({
                    "Name": ["Ave Sentiment"],
                    "Description": [round(sum(sentiments2)/len(sentiments2),2)]
                })
        if vis_description_df is not None:
            visDF2 = vis_description_df
        else:
             visDF2 = visDF
        
        visnet(visNodes,visEdges,
        network_title = vis_network_title,
        description_df = visDF2,
        description_title = vis_description_title,
        writeHTML = vis_writeHTML,
        browserView= vis_browserView,
        min_default_node_size= vis_min_default_node_size,
        min_default_edge_width = vis_min_default_edge_width,
        maximum_display = vis_maximum_display) 
    pbar.update(1)
    pbar.close() # Clean up the bar from the terminal
    return topicDFnoun, topicDFcomm,edgesDF1,edgesDF_global, edges_viz,nodeDegreeDF, visNodes, visEdges,similarNodesDF
    
def visTopic(edges:List[list], vis_network_title = 'Atsaniik',
            vis_description_df = None,
            vis_description_title = "Introduce your network",
            vis_writeHTML = None,
            vis_browserView= False,
            vis_min_default_node_size= 0,
            vis_min_default_edge_width = 0,
            vis_maximum_display = 100,colorNOUN = 'skyblue', shapeNOUN= 'box',
             colorADJ = 'gold', shapeADJ = 'triangle',
             colorVERB = 'pink', shapeVERB = 'square',
             colorADV ='brown', shapeADV = 'star',
             zeroWeightC = 'black',posWeightC ='green',negWeightC ='red',):
    """visualize the topic based on the aavn weighted edges + generatation of sentiment of edge

    Args:
        edges (list): weighted edges 
        vis_network_title (str, optional): _description_. Defaults to 'Atsaniik'.
        vis_description_df (_type_, optional): _description_. Defaults to None.
        vis_description_title (str, optional): _description_. Defaults to "Introduce your network".
        vis_writeHTML (_type_, optional): _description_. Defaults to None.
        vis_browserView (bool, optional): _description_. Defaults to False.
        vis_min_default_node_size (int, optional): _description_. Defaults to 0.
        vis_min_default_edge_width (int, optional): _description_. Defaults to 0.
        vis_maximum_display (int, optional): _description_. Defaults to 100.
        colorNOUN (str, optional): _description_. Defaults to 'skyblue'.
        shapeNOUN (str, optional): _description_. Defaults to 'box'.
        colorADJ (str, optional): _description_. Defaults to 'gold'.
        shapeADJ (str, optional): _description_. Defaults to 'triangle'.
        colorVERB (str, optional): _description_. Defaults to 'pink'.
        shapeVERB (str, optional): _description_. Defaults to 'square'.
        colorADV (str, optional): _description_. Defaults to 'brown'.
        shapeADV (str, optional): _description_. Defaults to 'star'.
        zeroWeightC (str, optional): _description_. Defaults to 'black'.
        posWeightC (str, optional): _description_. Defaults to 'green'.
        negWeightC (str, optional): _description_. Defaults to 'red'.

    Returns:
        visnet nodes,visnet edges and nodedegree: visNodesTopic, visEdgesTopic,nodeDegreeDF
    """
    if len(edges) < 1:
        raise ValueError("topic edges must have at least 1 edge")
    visNodesTopic, visEdgesTopic,nodeDegreeDF =preVisnet(edges,colorNOUN = colorNOUN, shapeNOUN= shapeNOUN,
             colorADJ = colorADJ, shapeADJ = shapeADJ,
             colorVERB = colorVERB, shapeVERB = shapeVERB,
             colorADV =colorADV, shapeADV = shapeADV,
             zeroWeightC = zeroWeightC,posWeightC =posWeightC,negWeightC =negWeightC)
    sentiments = []
    for edge in visEdgesTopic:
        sentiments.append (float(edge['title'].split('-')[-1]))
    
    visDF = pd.DataFrame({
                "Name": ["Nodes amount", "Edges amount","Ave Sentiment"],
                "Description": [len((visNodesTopic)), len((visEdgesTopic)),round(sum(sentiments)/len(sentiments),2)]
            })
    
    #visADD = rf"C:\PhDDocs\pythonPhD\myPYPI\kptopic2\topic_visual/topic_{topic}.html"

    if vis_description_df:
        visDF2 = vis_description_df
    else:
        visDF2 = visDF
    if vis_writeHTML:
        vis_writeHTML = vis_writeHTML
    else:
        time.sleep(1)
        now1 = datetime.now()
        now1 = now1.strftime("%Y_%m_%d_%H_%M_%S")
        vis_writeHTML = f"topic_{now1}.html"
    
    visnet(visNodesTopic,visEdgesTopic,
    network_title = vis_network_title,
    description_df = visDF2,
    description_title = vis_description_title,
    writeHTML = vis_writeHTML,
    browserView= vis_browserView,
    min_default_node_size= vis_min_default_node_size,
    min_default_edge_width = vis_min_default_edge_width,
    maximum_display = vis_maximum_display) 
    
    return visNodesTopic, visEdgesTopic,nodeDegreeDF



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

def visDIST(data,color='skyblue',edgecolor='black', alpha = 0.7, bins = 30, title = 'sentiment distribution', figsize = (10,6),saveADD = 'distribution_plot.png'):
    """visulize the distribtion of list of numbers 

    Args:
        data (list): a list of numbers 
        color (str, optional): _description_. Defaults to 'skyblue'.
        edgecolor (str, optional): _description_. Defaults to 'black'.
        alpha (float, optional): _description_. Defaults to 0.7.
        bins (int, optional): _description_. Defaults to 30.
        title (str, optional): _description_. Defaults to 'sentiment distribution'.
        figsize (tuple, optional): _description_. Defaults to (10,6).
        saveADD (str, optional): _description_. Defaults to 'distribution_plot.png'.
    """
    
    
    #data = edgesDF1.round(2).sentiment.tolist()

    # 2. Calculate mean and standard deviation
    mean = float(np.mean(data))
    std_dev = float(np.std(data))

    # 3. Create the plot
    plt.figure(figsize=(10, 6))
    sns.histplot(data, kde=True, color=color, edgecolor= edgecolor, alpha=alpha,bins=bins,)

    # 4. Add vertical lines for mean and standard deviation
    # Solid line for the mean
    plt.axvline(mean, color='red', linestyle='-', linewidth=2, 
                label=rf'Mean ($\mu$): {mean:.2f}')

    # Dashed lines for standard deviations
    plt.axvline(mean - std_dev, color='green', linestyle='--', linewidth=1.5, 
                label=rf'1 Std Dev ($\sigma$): {std_dev:.2f}')
    plt.axvline(mean + std_dev, color='green', linestyle='--', linewidth=1.5)

    # 5. Add labels and legend
    plt.title(title, fontsize=16)
    plt.xlabel('Value', fontsize=12)
    plt.ylabel('Frequency', fontsize=12)
    plt.legend()

    # 6. Save the plot
    plt.savefig(saveADD)



if __name__ == "__main__":
    
    texts = ['I love you, I would not come to finland this year😊 visit us at uef.fi ',
             'I do not love you, I can go to school now, contact us peng@uef.fi , we will have great food',
             'this is extremely good food, you must try it',
             'Finland has a lot delicious food and drinks, highly recommended',
             'I do not recommend🤷‍♂️ you eat the food',
             'Finland is beautiful country','I do not love you']
    
    edgesDF,eueDF = semanticX(texts, verbPhrase= True,coreference= False,corefPath=None,)
    topicDFnoun, topicDFcomm,edgesDF1,edgesDF_global, edges_viz,nodeDegreeDF, visNodes, visEdges,similarNodesDF = netKPT(edgesDF,colorNOUN='skyblue',kCliqueCOMMtopic=2,removeNodes = [],minSimilar=.96,visualizeNET=True)
    
  
    with pd.ExcelWriter(r"C:\PhDDocs\pythonPhD\myPYPI\kptopic2/kpt_edges.xlsx") as writer:
        topicDFnoun.to_excel(writer, sheet_name='topicNOUN',index=False)
        topicDFcomm.to_excel(writer,sheet_name='topicCOMM',index=False)
        edges_viz.to_excel(writer,sheet_name='edges_vis',index=False)
        #edgesDF_global.to_excel(writer,sheet_name='cleaned weighted',index= False)
        edgesDF1.to_excel(writer,sheet_name='cleaned edges',index= False)
        edgesDF.to_excel(writer, sheet_name='original edges', index= False)
        nodeDegreeDF.to_excel(writer,sheet_name ='node_degree')
        eueDF.to_excel(writer, sheet_name='eue', index= False)
        similarNodesDF.to_excel(writer,sheet_name='simNodes')
    # AI Prompt to generate topic:
    """You are given a central keyword and a list of related words with numerical weights representing their associations. Positive weights indicate positive associations, negative weights indicate negative associations, and zero weight indicates neutral association. Using this information, write one coherent topic sentence about the central keyword that naturally reflects both the positive and negative associations. Include the most important related words, convey the overall sentiment, and ensure the sentence reads smoothly as if written by a human."

    Example Input:
    Central word: food
    Edges: {'Finland_2noun': 1, 'delicious_2adj': 1, 'drink_2noun': 1, 'good_2adj': 1, 'lot_2noun': 1, 'have_2verb': 1, 'recommend_2verb': 0, 'eat_2verb': -1, '\u200d_2noun': -1, '♂_2noun': -1}

    Example Output (Topic Sentence):
    "In Finland, the food is generally good and delicious, with plenty of drinks available, though some dishes may not be enjoyable or suitable for everyone."""



import requests
import base64
import json, os
from IPython.display import HTML, display
import pandas as pd
import webbrowser
import random

def base64_from_url(image_url):
    """
    Downloads an image from a URL and returns its Base64 encoded string.

    Args:
        image_url (str): The URL of the image.

    Returns:
        str: The Base64 encoded string of the image, prefixed with data URI,
             or None if an error occurs.
    """
    try:
        response = requests.get(image_url, stream=True)
        response.raise_for_status()

        content_type = response.headers['Content-Type']
        if not content_type.startswith('image/'):
            print(f"Warning: URL does not point to an image. Content-Type: {content_type}")
            return None

        image_bytes = response.content
        encoded_string = base64.b64encode(image_bytes).decode('utf-8')
        return f"data:{content_type};base64,{encoded_string}"

    except requests.exceptions.RequestException as e:
        print(f"Error downloading image from {image_url}: {e}")
        return None
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        return None

def base64_from_loc(image_local):
    """ turn local image to base64 for vis.js diagram for an image node 

    Args:
        image_local (_type_): local image address 
    """
    try:
        with open(image_local, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read()).decode("utf-8")
            base64_image_local = f"data:image/png;base64,{encoded_string}"
            return base64_image_local
    except FileNotFoundError:
        print("Error: Image file not found. Using placeholder image.")
        return None      

def visnet(nodes, edges, network_title='Atsaniik', description_df=None, description_title="Introduce your network", writeHTML="network_visualization.html", browserView=False, min_default_node_size=0, min_default_edge_width=0, maximum_display=100):
    """ visualize your network with a save button to export selected nodes and edges to JSON files
    Args:
        nodes(list): list of nodes dict e.g. {"id": 1, "label": "Start", "size": 30, "color": "red-spain", "shape": "dot-america", "title": "starting the point"}
        edges(list): list of edges dict e.g. {"from": 1, "to": 3, "width": 4, "color": {"color": "#C70039-europe"}, "arrows": "to", "title": "from 1 to 3"}
        description_df(dataframe): a dataframe with columns "Name", "Description"
        description_title(str): describe your network in short
        writeHTML(str): output HTML file name
        browserView(bool): whether to open the HTML in a browser
        min_default_node_size(float): minimum node size for initial display
        min_default_edge_width(float): minimum edge width for initial display
        maximum_display(int): maximum number of nodes to display at any time
    """
    defaults_node = {"label": lambda node: str(node['id']), "size": 1, "color": "gray", "shape": "dot", "title": lambda node: str(node['id'])}

    nodeIDchecks = []
    for node in nodes:
        nodeIDcheck = node['id']
        nodeIDchecks.append(nodeIDcheck)
        for key, default_value in defaults_node.items():
            if key not in node:
                # If the key is lon or node_hover, use the lambda to compute the value
                node[key] = default_value(node) if callable(default_value) else default_value
    #print(nodes)


    defaults_edge = {"width": 1, "color": {"color": "gray"}, "title": lambda edge: str(edge['from'])+ ' - ' + str(edge['to'] )}
    edgeIDchecks = []
    for edge in edges:
        edgeIDcheck = [edge['from'], edge['to']]
        missingNodes = [x for x in edgeIDcheck if x not in nodeIDchecks]
        if len(missingNodes)>0:
            print(f'edge {edge['from']}-{edge['to']} missing nodes description:  {missingNodes}')
        edgeIDchecks.extend(edgeIDcheck)
        for key, default_value in defaults_edge.items():
            if key not in edge:
                # If the key is lon or node_hover, use the lambda to compute the value
                edge[key] = default_value(edge) if callable(default_value) else default_value
    #print(edges)
  
  
    
    def extract_base_value(full_value, is_shape=False):
        if "-" in full_value:
            return full_value.split("-")[0]
        return full_value

    node_colors = [
        node["color"]["background"] if isinstance(node["color"], dict) else node["color"]
        for node in nodes if "color" in node
               ]


    edge_colors = [edge["color"]["color"] for edge in edges if "color" in edge and "color" in edge["color"]]

    color_bases = {}
    for color in node_colors + edge_colors:
        base = extract_base_value(color)
        if base in color_bases:
            color_bases[base].add(str(color))
        else:
            color_bases[base] = {str(color)}
            
            
    for base, variants in color_bases.items():
        if len(variants) > 1:
            print(f"Error: Multiple extra labels detected for base color '{base}': {', '.join(variants)}. Execution stopped.")
            return

    shape_bases = {}
    icon_shapes = {}
    image_shapes = {}
    for node in nodes:
            if "shape" in node:
                full_shape = node["shape"]
                base = extract_base_value(full_shape, is_shape=True)
                if base in shape_bases:
                    shape_bases[base].add(full_shape)
                else:
                    shape_bases[base] = {full_shape}
                # Track icon and image consistency
                if base == "icon" and "icon" in node and "code" in node["icon"]:
                    shape_label = full_shape.split("-")[1] if "-" in full_shape else full_shape
                    if node["icon"]["code"] in icon_shapes:
                        if shape_label not in icon_shapes[node["icon"]["code"]]:
                            icon_shapes[node["icon"]["code"]].add(shape_label)
                        else:
                            continue
                    else:
                        icon_shapes[node["icon"]["code"]] = {shape_label}
                elif base == "image" and "image" in node:
                    shape_label = full_shape.split("-")[1] if "-" in full_shape else full_shape
                    image_value = str(node["image"])
                    if image_value in image_shapes:
                        if shape_label not in image_shapes[image_value]:
                            image_shapes[image_value].add(shape_label)
                        else:
                            continue
                    else:
                        image_shapes[image_value] = {shape_label}

    for code, shapes in icon_shapes.items():
        if len(shapes) > 1:
            print(f"Error: Inconsistent icon code '{code}' used with different shape names: {', '.join(shapes)}. Execution stopped.")
            return

    for image_value, shapes in image_shapes.items():
        if len(shapes) > 1:
            print(f"Error: Inconsistent image name '{image_value}' used with different shape names: {', '.join(shapes)}. Execution stopped.")
            return

    for base, variants in shape_bases.items():
        if len(variants) > 1 and base not in ["icon", "image"]:
            print(f"Error: Multiple extra labels detected for base shape '{base}': {', '.join(variants)}. Execution stopped.")
            return

    if description_df is None:
        df = pd.DataFrame({
            "Name": ["Nodes amount", "Edges amount"],
            "Description": [len(nodes), len(edges)]
        })
    else:
        df = description_df

    table_rows = "".join([f'<tr><td>{row["Name"]}</td><td>{row["Description"]}</td></tr>' for _, row in df.iterrows()])

    all_nodes = sorted(nodes, key=lambda x: x["size"], reverse=True)
    unique_sizes = len(set(node["size"] for node in all_nodes))
    if unique_sizes == 1:
        random.shuffle(all_nodes)
    initial_nodes = all_nodes[:maximum_display]
    initial_node_ids = set(node["id"] for node in initial_nodes)

    initial_edges = [
        edge for edge in edges
        if edge["from"] in initial_node_ids and edge["to"] in initial_node_ids and 
        (min_default_edge_width is None or edge["width"] >= min_default_edge_width)
    ]

    min_size = round(min(node["size"] for node in all_nodes), 2) if all_nodes else 0.00
    max_size = round(max(node["size"] for node in all_nodes), 2) if all_nodes else 0.00
    min_width = round(min(edge["width"] for edge in edges), 2) if edges else 0.00
    max_width = round(max(edge["width"] for edge in edges), 2) if edges else 0.00

    """def extract_base_value(full_value, is_shape=False):
        if isinstance(full_value, dict):
            full_value = full_value.get('background', '') if is_shape else full_value.get('color', '') if not is_shape else ''
        if "-" in full_value:
            return full_value.split("-")[0]
        return full_value"""
    
    
    

    colors = sorted(set(
        node["color"]["background"] if isinstance(node["color"], dict) else node["color"]
        for node in all_nodes
      ))
    shapes = sorted(set(node["shape"] for node in nodes))
    labels = sorted(set(node["label"] for node in all_nodes))
    titles = sorted(set(node.get("title", "") for node in all_nodes if "title" in node))
    edge_colors = sorted(set(edge["color"]["color"] for edge in edges))

    for node in all_nodes:
        full_shape = node["shape"]
        base_shape = extract_base_value(full_shape, is_shape=True)
        node["shape_label"] = full_shape
        node["shape"] = base_shape
        if isinstance(node["color"], dict):
            color_base = extract_base_value(node["color"].get('background', ''))
        else:
            color_base = extract_base_value(node["color"])
        node["color"] = {"background": color_base, "border": node["color"].get('border', 'gray') if isinstance(node["color"], dict) else 'gray'}
        
        
    for edge in edges:
        edge["color"]["color"] = extract_base_value(edge["color"]["color"])

    color_options = "\n".join([f'            <option value="{color}">{color}</option>' for color in colors])
    shape_options = "\n".join([f'            <option value="{shape}">{shape}</option>' for shape in shapes])
    title_options = "\n".join([f'            <option value="{title}">{title}</option>' for title in titles if title])
    edge_color_options = "\n".join([f'            <option value="{color}">{color}</option>' for color in edge_colors])
    max_display_options = "\n".join([f'            <option value="{val}"{" selected" if val == maximum_display else ""}>{val}</option>' for val in [1,5,10,20,30,50,60,70,80,90,100,108,150,200,300,400,500,1000,2000,5000,10000,20000]])

    nodes_json = json.dumps(initial_nodes, ensure_ascii=False)
    all_nodes_json = json.dumps(all_nodes, ensure_ascii=False)
    all_edges_json = json.dumps(edges, ensure_ascii=False)
    edges_json = json.dumps(initial_edges, ensure_ascii=False)

    drag_script = """
    <script type="text/javascript">
        var isDragging = false;
        var currentX;
        var currentY;
        var initialX;
        var initialY;
        var titleElement = document.getElementById('diagram-title');

        function startDragging(e) {
            initialX = e.clientX - currentX;
            initialY = e.clientY - currentY;
            isDragging = true;
            titleElement.style.cursor = 'grabbing';
        }

        function stopDragging() {
            isDragging = false;
            titleElement.style.cursor = 'move';
        }

        function dragTitle(e) {
            if (isDragging) {
                e.preventDefault();
                currentX = e.clientX - initialX;
                currentY = e.clientY - initialY;
                titleElement.style.left = currentX + 'px';
                titleElement.style.top = currentY + 'px';
            }
        }

        if (titleElement) {
            currentX = parseInt(window.getComputedStyle(titleElement).left) || 50;
            currentY = parseInt(window.getComputedStyle(titleElement).top) || 10;
            titleElement.addEventListener('mousedown', startDragging);
            document.addEventListener('mousemove', dragTitle);
            document.addEventListener('mouseup', stopDragging);
            titleElement.style.userSelect = 'none';
        }
    </script>
    """

    script_code = """
    <script type="text/javascript">
    var allNodes = {all_nodes_json};
    var allEdges = {all_edges_json};
    var originalNodes = {nodes_json};
    var originalEdges = {edges_json};
    var nodes = new vis.DataSet(originalNodes);
    var edges = new vis.DataSet(originalEdges);
    var lastZoomCenter = {{ x: 0, y: 0 }};
    var lastScale = 1;
    var uniqueLabels = [...new Set(allNodes.map(node => node.label).filter(label => label))].sort();
    var previousNodes = originalNodes;
    var previousEdges = originalEdges;

    var container = document.getElementById('mynetwork');
    if (!container) {{
        console.error("Container #mynetwork not found!");
    }}
    var loadingBar = document.getElementById('loading-bar');
    var loadingProgress = document.getElementById('loading-progress');
    var loadingPercentage = document.getElementById('loading-percentage');

    var data = {{
        nodes: nodes,
        edges: edges
    }};
    var options = {{
        nodes: {{
            shape: 'dot',
            size: 20,
            font: {{ size: 14, color: '#333333' }},
            borderWidth: 2,
            color: {{
                background: '#97C2E6',
                border: '#2B7CE9',
                highlight: {{ background: '#D2E5FF', border: '#2B7CE9' }},
                hover: {{ background: '#D2E5FF', border: '#2B7CE9' }}
            }}
        }},
        edges: {{
            width: 1,
            color: {{ color: '#848484', highlight: '#848484', hover: '#848484', inherit: 'from', opacity: 0.8 }},
            smooth: {{ enabled: true, type: "dynamic" }},
            font: {{ size: 12 }}
        }},
        physics: {{
            enabled: true,
            stabilization: {{
                enabled: true,
                iterations: 1000,
                updateInterval: 50
            }},
            barnesHut: {{
                gravitationalConstant: -2000,
                centralGravity: 0.3,
                springLength: 95,
                springConstant: 0.04,
                damping: 0.09,
                avoidOverlap: 0
            }},
            solver: 'barnesHut'
        }},
        configure: {{
            enabled: true,
            filter: 'physics',
            showButton: true,
            container: document.getElementById('physics-config')
        }}
    }};

    container.style.opacity = '0';

    var network = new vis.Network(container, data, options);
    if (!network) {{
        console.error("Failed to initialize vis.Network!");
    }}

    network.on("stabilizationProgress", function(params) {{
        var progress = (params.iterations / params.total) * 100;
        if (loadingProgress && loadingPercentage) {{
            loadingProgress.style.width = progress + '%';
            loadingPercentage.textContent = Math.round(progress) + '%';
        }}
    }});

    network.on("stabilizationIterationsDone", function() {{
        if (loadingBar) {{
            loadingBar.style.display = 'none';
        }}
        container.style.opacity = '1';
        network.setOptions({{ physics: false }});
        updateNodes();
    }});

    function getSelectedOptions(selectElement) {{
        return Array.from(selectElement.selectedOptions).map(option => option.value);
    }}

    function updateSuggestions() {{
        var input = document.getElementById('label-search').value;
        var suggestionsDiv = document.getElementById('label-suggestions');
        suggestionsDiv.innerHTML = '';
        suggestionsDiv.style.display = 'none';

        if (input.trim() !== '') {{
            var terms = input.split(',').map(term => term.trim().toLowerCase());
            var lastTerm = terms[terms.length - 1];
            if (lastTerm) {{
                var matches = uniqueLabels.filter(label => 
                    label.toLowerCase().includes(lastTerm)
                );
                if (matches.length > 0) {{
                    suggestionsDiv.style.display = 'block';
                    matches.forEach(label => {{
                        var div = document.createElement('div');
                        div.className = 'suggestion-item';
                        div.textContent = label;
                        div.onclick = function() {{
                            var currentTerms = input.split(',').slice(0, -1);
                            currentTerms.push(label);
                            document.getElementById('label-search').value = currentTerms.join(',') + (currentTerms.length > 0 ? ',' : '');
                            suggestionsDiv.innerHTML = '';
                            suggestionsDiv.style.display = 'none';
                            updateNodes();
                        }};
                        suggestionsDiv.appendChild(div);
                    }});
                }}
            }}
        }} else {{
            updateNodes();
        }}
    }}

    function saveSelection() {{
        var currentNodes = nodes.get();
        var currentEdges = edges.get();

        // Create JSON strings
        var nodesJson = JSON.stringify(currentNodes, null, 2);
        var edgesJson = JSON.stringify(currentEdges, null, 2);

        // Create Blob objects for download
        var nodesBlob = new Blob([nodesJson], {{ type: 'application/json' }});
        var edgesBlob = new Blob([edgesJson], {{ type: 'application/json' }});

        // Create temporary URLs for the Blobs
        var nodesUrl = URL.createObjectURL(nodesBlob);
        var edgesUrl = URL.createObjectURL(edgesBlob);

        // Create download links
        var nodesLink = document.createElement('a');
        nodesLink.href = nodesUrl;
        nodesLink.download = 'nodes.json';
        document.body.appendChild(nodesLink);
        nodesLink.click();
        document.body.removeChild(nodesLink);

        var edgesLink = document.createElement('a');
        edgesLink.href = edgesUrl;
        edgesLink.download = 'edges.json';
        document.body.appendChild(edgesLink);
        edgesLink.click();
        document.body.removeChild(edgesLink);

        // Clean up URLs
        URL.revokeObjectURL(nodesUrl);
        URL.revokeObjectURL(edgesUrl);

        // Log saved locations (simulating file paths for console output)
        console.log("Nodes saved to: nodes.json");
        console.log("Edges saved to: edges.json");
    }}

    function updateNodes(zoomCenter = null, zoomDirection = null) {{
        console.log("Updating nodes...");
        var maxDisplay = parseInt(document.getElementById('max-display-select').value) || {maximum_display};
        var colorSelect = getSelectedOptions(document.getElementById('color-select'));
        var shapeSelect = getSelectedOptions(document.getElementById('shape-select'));
        var sizeMin = parseFloat(document.getElementById('size-min').value) || 0;
        var sizeMax = parseFloat(document.getElementById('size-max').value) || Infinity;
        var labelSearch = document.getElementById('label-search').value;
        var titleSelect = getSelectedOptions(document.getElementById('title-select'));
        var widthMin = document.getElementById('width-min').value ? parseFloat(document.getElementById('width-min').value) : null;
        var widthMax = document.getElementById('width-max').value ? parseFloat(document.getElementById('width-max').value) : null;
        var edgeColorSelect = getSelectedOptions(document.getElementById('edge-color-select'));

        var filteredNodes = allNodes;
        var filteredEdges = allEdges;

        if (labelSearch.trim() !== '') {{
            previousNodes = nodes.get();
            previousEdges = edges.get();
            console.log("Saving previous state: ", previousNodes.length, "nodes,", previousEdges.length, "edges");

            var searchTerms = labelSearch.split(',').map(term => term.trim().toLowerCase()).filter(term => term !== '');
            console.log("Search terms:", searchTerms);
            var selectedNodes = allNodes.filter(function(node) {{
                return node.label && searchTerms.some(term => node.label.toLowerCase() === term);
            }});
            console.log("Selected nodes:", selectedNodes.map(node => node.label));
            var neighborIds = new Set();
            selectedNodes.forEach(function(node) {{
                neighborIds.add(node.id);
                allEdges.forEach(function(edge) {{
                    if (edge.from === node.id) {{
                        neighborIds.add(edge.to);
                    }}
                    if (edge.to === node.id) {{
                        neighborIds.add(edge.from);
                    }}
                }});
            }});
            console.log("Neighbor IDs:", Array.from(neighborIds));
            filteredNodes = allNodes.filter(function(node) {{
                return neighborIds.has(node.id);
            }});
            console.log("Filtered nodes:", filteredNodes.map(node => node.label));
        }} 

        var validNodeIds = new Set();
        if (widthMin !== null || widthMax !== null) {{
            filteredEdges = allEdges.filter(function(edge) {{
                var widthMatch = true;
                if (widthMin !== null) {{
                    widthMatch = widthMatch && edge.width >= widthMin;
                }}
                if (widthMax !== null) {{
                    widthMatch = widthMatch && edge.width <= widthMax;
                }}
                return widthMatch;
            }});
            if (widthMin !== null && widthMin > 0) {{
                filteredEdges.forEach(function(edge) {{
                    validNodeIds.add(edge.from);
                    validNodeIds.add(edge.to);
                }});
                filteredNodes = filteredNodes.filter(function(node) {{
                    return validNodeIds.has(node.id) || filteredEdges.some(edge => edge.from === node.id || edge.to === node.id);
                }});
            }}
        }}

        filteredNodes = filteredNodes.filter(function(node) {{
            var fullColor = typeof node.color === 'string' ? node.color : node.color.background;
            var fullShape = node.shape_label || node.shape;
            var colorMatch = colorSelect.includes('all') || colorSelect.some(cs => fullColor === cs || fullColor.startsWith(cs.split('-')[0]));
            var shapeMatch = shapeSelect.includes('all') || (shapeSelect.length === 1 && shapeSelect[0] !== 'all' && fullShape === shapeSelect[0]);
            var sizeMatch = node.size >= sizeMin && node.size <= sizeMax;
            var titleMatch = titleSelect.includes('all') || (node.title && titleSelect.includes(node.title));
            return colorMatch && shapeMatch && sizeMatch && titleMatch;
        }});

        filteredNodes.sort((a, b) => b.size - a.size);

        filteredNodes = filteredNodes.slice(0, maxDisplay);

        filteredEdges = allEdges.filter(function(edge) {{
            var fullEdgeColor = edge.color.color;
            var colorMatch = edgeColorSelect.includes('all') || edgeColorSelect.some(function(ecs) {{
                return fullEdgeColor === ecs || fullEdgeColor.startsWith(ecs.split('-')[0]);
            }});

            var widthMatch = true;
            if (widthMin !== null) {{
                widthMatch = widthMatch && edge.width >= widthMin;
            }}
            if (widthMax !== null) {{
                widthMatch = widthMatch && edge.width <= widthMax;
            }}

            var nodeMatch = filteredNodes.some(function(n) {{
                return n.id === edge.from;
            }}) && filteredNodes.some(function(n) {{
                return n.id === edge.to;
            }});
            return colorMatch && widthMatch && nodeMatch;
        }});

        document.getElementById('node-count').textContent = filteredNodes.length;
        document.getElementById('edge-count').textContent = filteredEdges.length;

        nodes.clear();
        nodes.add(filteredNodes);
        edges.clear();
        edges.add(filteredEdges);
    }}

    function togglePanel(panelId) {{
        console.log("Toggling panel: " + panelId);
        var panel = document.getElementById(panelId);
        if (panel) {{
            panel.classList.toggle('panel-hidden');
        }} else {{
            console.error("Panel " + panelId + " not found!");
        }}
    }}

    network.on("zoom", function(params) {{
        var scale = params.scale;
        if (scale > lastScale) {{
            var pointerDOM = params.pointer.DOM;
            var pointerCanvas = network.DOMtoCanvas(pointerDOM);
            lastZoomCenter = pointerCanvas;
            updateNodes(lastZoomCenter, 'in');
        }} else if (scale < lastScale) {{
            updateNodes(null, 'out');
        }}
        lastScale = scale;
    }});

    // Add Enter key handler for label-search input
    document.getElementById('label-search').addEventListener('keydown', function(event) {{
        if (event.key === 'Enter') {{
            document.getElementById('label-suggestions').style.display = 'none';
            updateNodes();
        }}
    }});

    updateNodes();
    </script>
    """.format(
        all_nodes_json=all_nodes_json,
        all_edges_json=all_edges_json,
        nodes_json=nodes_json,
        edges_json=edges_json,
        maximum_display=maximum_display
    )

    html_code = f"""
    <html>
    <head>
    <script type="text/javascript" src="https://unpkg.com/vis-network/standalone/umd/vis-network.min.js"></script>
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.5.2/css/all.min.css" integrity="sha512-SnH5WK+bZxgPHs44uWIX+LLJAJ9/2PkPKZ5QiAj6Ta86w+fsb2TkcmfRyVX3pBnMFcV7oQPJkl9QevSCWr3W6A==" crossorigin="anonymous" referrerpolicy="no-referrer" />
    <style type="text/css">
        body {{
            font-family: 'Inter', sans-serif;
            display: flex;
            flex-direction: column;
            justify-content: center;
            align-items: center;
            min-height: 100vh;
            margin: 0;
            background-color: #e0e0e0;
            position: relative;
        }}
        #mynetwork {{
            width: 100vw;
            height: 100vh;
            border: 1px solid lightgray;
            background-color: #f9f9f9;
            border-radius: 8px;
            box-shadow: 0 4px 8px rgba(0,0,0,0.1);
            transition: opacity 0.5s ease;
        }}
        #diagram-title {{
            position: absolute;
            top: 10px;
            left: 50px;
            z-index: 100;
            background-color: rgba(255, 255, 255, 0.8);
            padding: 10px 15px;
            border-radius: 5px;
            box-shadow: 0 2px 4px rgba(0,0,0,0.1);
            font-size: 24px;
            color: #333;
            cursor: move;
            user-select: none;
        }}
        #loading-bar {{
            position: absolute;
            top: 50%;
            left: 50%;
            transform: translate(-50%, -50%);
            width: 300px;
            height: 30px;
            background-color: #ddd;
            border-radius: 15px;
            overflow: hidden;
            z-index: 150;
            box-shadow: 0 2px 4px rgba(0,0,0,0.2);
        }}
        #loading-progress {{
            width: 0%;
            height: 100%;
            background-color: #4CAF50;
            transition: width 0.3s ease;
        }}
        #loading-percentage {{
            position: absolute;
            top: 50%;
            left: 50%;
            transform: translate(-50%, -50%);
            font-size: 16px;
            font-weight: bold;
            color: #333;
            z-index: 151;
        }}
        #config-container {{
            position: absolute;
            bottom: 20px;
            right: 20px;
            z-index: 100;
            display: flex;
            flex-direction: column;
            gap: 10px;
            max-height: calc(100vh - 40px);
            max-width: 20vw;
            overflow-y: auto;
            padding: 10px;
        }}
        #description-container {{
            position: absolute;
            bottom: 20px;
            left: 20px;
            z-index: 100;
            display: flex;
            flex-direction: column;
            gap: 10px;
            max-height: calc(100vh - 40px);
            max-width: 20vw;
            overflow-y: auto;
            padding: 10px;
        }}
        #save-button-container {{
            position: absolute;
            bottom: 20px;
            left: 50%;
            transform: translateX(-50%);
            z-index: 100;
            display: flex;
            justify-content: center;
        }}
        #save-selection {{
            background-color: white;
            color: #333333;
            padding: 10px 20px;
            border: none;
            border-radius: 5px;
            cursor: pointer;
            font-size: 16px;
            font-weight: bold;
            box-shadow: 0 2px 4px rgba(0,0,0,0.2);
            transition: background-color 0.3s ease;
        }}
        #save-selection:hover {{
            background-color: #45a049;
        }}
        #node-config, #physics-config, #description-table {{
            background-color: white;
            padding: 10px;
            border-radius: 5px;
            border: 1px solid #ccc;
            max-width: 100%;
            overflow-x: auto;
        }}
        #node-config .filter-group, #description-table .table-group {{
            margin-bottom: 20px;
        }}
        #node-config .filter-group label, #description-table .table-group label {{
            display: block;
            font-weight: bold;
            margin-bottom: 5px;
            
        }}
        .label-search {{ 
            color: #0066cc; /* Added color for Node Label Search label */
            display: block;              /* ensures it takes full width */
            text-align: center;          /* centers the text */
            font-size: 1.2em;            /* makes the label larger */
            font-weight: bold;           /* optional: makes it stand out */
            margin-bottom: 10px;         /* space between label and input */
            
        }}
        #label-search {{
                        display: block;
                        margin: 0 auto;
                        width: 80%;
                        height: 40px;            /* increases height */
                        padding: 10px 12px;      /* adds internal space */
                        font-size: 1.1em;        /* makes text slightly larger */
                        border: 1px solid #ccc;
                        border-radius: 6px;      /* optional rounded corners */
                        box-sizing: border-box;
                    }}
        #node-config select, #node-config input[type="number"], #node-config input[type="text"], #description-table table {{
            width: 100%;
            padding: 5px;
            border-radius: 3px;
            border: 1px solid #ccc;
            box-sizing: border-box;
        }}
        #node-config .size-inputs {{
            display: flex;
            gap: 5px;
        }}
        #node-config .size-inputs input[type="number"] {{
            flex: 1;
        }}
        .legend {{
            cursor: pointer;
            background-color: rgba(255, 255, 255, 0.9);
            padding: 8px 12px;
            border-radius: 5px;
            box-shadow: 0 2px 4px rgba(0,0,0,0.1);
            font-size: 14px;
            font-weight: bold;
            color: #333;
            text-align: center;
        }}
        .panel-hidden {{
            display: none;
        }}
        #node-config p {{
            font-size: 12px;
            color: #555;
            margin: 5px 0;
        }}
        .vis-icon {{
            font-family: "Font Awesome 5 Free" !important;
            font-weight: 900;
        }}
        #description-table table {{
            border-collapse: collapse;
            width: 100%;
        }}
        #description-table th, #description-table td {{
            border: 1px solid #ddd;
            padding: 8px;
            text-align: left;
        }}
        #description-table th {{
            background-color: #f2f2f2;
        }}
        #color-select, #shape-select, #title-select, #edge-color-select {{
            height: 100px;
        }}
        #node-edge-count {{
            background-color: rgba(255, 255, 255, 0.9);
            padding: 8px 12px;
            border-radius: 5px;
            box-shadow: 0 2px 4px rgba(0,0,0,0.1);
            font-size: 14px;
            font-weight: bold;
            color: #333;
            text-align: center;
        }}
        .suggestions {{
            position: absolute;
            background-color: white;
            border: 1px solid #ccc;
            border-radius: 3px;
            max-height: 150px;
            overflow-y: auto;
            width: 100%;
            z-index: 200;
            box-shadow: 0 2px 4px rgba(0,0,0,0.1);
            display: none;
        }}
        .suggestion-item {{
            padding: 5px 10px;
            cursor: pointer;
        }}
        .suggestion-item:hover {{
            background-color: #f0f0f0;
        }}
    </style>
    </head>
    <body>
    <h2 id="diagram-title">{network_title}</h2>
    <div id="mynetwork"></div>
    <div id="loading-bar">
        <div id="loading-progress"></div>
        <span id="loading-percentage">0%</span>
    </div>
    <div id="config-container">
        <div class="legend" onclick="togglePanel('node-config')">Node & Edge Selection</div>
        <div id="node-config" class="panel-hidden">
            <div class="filter-group">
                <label class="label-search">Node Label Search</label>
                <input type="text" id="label-search" placeholder="Enter labels, e.g., Node1,Node2" oninput="updateSuggestions()">
                <div id="label-suggestions" class="suggestions"></div>
            </div>
            <div class="filter-group">
                <label>Maximum Nodes to Display</label>
                <select id="max-display-select" onchange="updateNodes()">
                    {max_display_options}
                </select>
            </div>
            <div class="filter-group">
                <label>Node Color</label>
                <select id="color-select" multiple onchange="updateNodes()">
                    <option value="all" selected>All</option>
                    {color_options}
                </select>
            </div>
            <div class="filter-group">
                <label>Node Shape</label>
                <select id="shape-select" multiple onchange="updateNodes()">
                    <option value="all" selected>All</option>
                    {shape_options}
                </select>
            </div>
            <div class="filter-group">
                <label>Node Size ({min_size}-{max_size})</label>
                <div class="size-inputs">
                    <input type="number" id="size-min" value="{min_default_node_size}" placeholder="Min" min="0" oninput="updateNodes()">
                    <input type="number" id="size-max" placeholder="Max" min="0" oninput="updateNodes()">
                </div>
            </div>

            <div class="filter-group">
                <label>Node Title (Class)</label>
                <select id="title-select" multiple onchange="updateNodes()">
                    <option value="all" selected>All</option>
                    {title_options}
                </select>
            </div>
            <div class="filter-group">
                <label>Edge Width ({min_width}-{max_width})</label>
                <div class="size-inputs">
                    <input type="number" id="width-min" value="{min_default_edge_width}" placeholder="min" min="0" oninput="updateNodes()">
                    <input type="number" id="width-max" placeholder="Max" min="0" oninput="updateNodes()">
                </div>
            </div>
            <div class="filter-group">
                <label>Edge Color</label>
                <select id="edge-color-select" multiple onchange="updateNodes()">
                    <option value="all" selected>All</option>
                    {edge_color_options}
                </select>
            </div>
        </div>
        <div class="legend" onclick="togglePanel('physics-config')">Physics</div>
        <div id="physics-config" class="panel-hidden"></div>
    </div>
    <div id="description-container">
        <div id="node-edge-count" class="legend">
            Nodes: <span id="node-count">0</span>, Edges: <span id="edge-count">0</span>
        </div>
        <div class="legend" onclick="togglePanel('description-table')">Diagram Description</div>
        <div id="description-table" class="panel-hidden">
            <p>{description_title}</p>
            <div class="table-group">
                <table>
                    <thead>
                        <tr>
                            <th>Name</th>
                            <th>Description</th>
                        </tr>
                    </thead>
                    <tbody>
                    {table_rows}
                    </tbody>
                </table>
            </div>
        </div>
    </div>
    <div id="save-button-container">
        <button id="save-selection" onclick="saveSelection()">Save Selection</button>
    </div>
    {drag_script}
    {script_code}
    </body>
    </html>
    """

    if writeHTML is None:
        html_add = "network_visualization.html"
    else:
        html_add = writeHTML 
    
    with open(html_add, "w", encoding="utf-8") as file:
        file.write(html_code)
    base_path = os.path.dirname(os.path.abspath(__file__))

    if browserView:
        webbrowser.open(html_add)

    print(f"HTML file has been created: {base_path}/{writeHTML}")
    return None


if __name__ == "__main__":

    
    gipuzkoa_path = "https://upload.wikimedia.org/wikipedia/commons/thumb/a/a1/Escudo_de_Donostia.svg/800px-Escudo_de_Donostia.svg.png"
    turku_path = "https://upload.wikimedia.org/wikipedia/commons/thumb/d/d7/Turku.vaakuna.svg/800px-Turku.vaakuna.svg.png"
    
    turku_path2 = r"C:\PhDDocs\pythonPhD\pyBase4All\hyperlinks_scrape\scrapper\network_visual\network_visual\Turku.vaakuna.png"
    gipuzkoa_path2 = r"C:\PhDDocs\pythonPhD\pyBase4All\hyperlinks_scrape\scrapper\network_visual\network_visual\Escudo_de_Donostia.png"
    
    turku_img64 = base64_from_loc(turku_path2)
    gipuzkoa_img64 = base64_from_loc(gipuzkoa_path2 )

    nodes = [
        {"id": 1, "label": "Start", "size": 9, "color": "red-spain", "shape": "dot", "title": "starting the point"},
        {"id": 2, "label": "Process A", "size": 25, "color": "#33FF57-france", "shape": "triangle", "title": "process one"},
        {"id": 3, "label": "Process B", "size": 40, "color": "#3357FF-france", "shape": "box-canada", "title": "process two"},
        {"id": 4, "label": "Critical", "size": 35, "color": "#FFBD33-japan", "shape": "star-asia", "title": "critical point"},
        {"id": 5, "label": "User", "size": 45, "color": "#8D33FF-germany", "shape": "icon-africa", 
            "icon": {"face": '"Font Awesome 5 Free"', "code": "\uf007", "size": 50, "color": "#8D33FF-germany"}},
        {"id": 6, "label": "Info A", "size": 48, "color": "#33B5E5-canada", "shape": "icon-africa",
            "icon": {"face": "Font Awesome 5 Brands", "code": "\uf007", "size": 40, "color": "#33B5E5-canada"}},
        {"id": 7, "label": "Image Node A ", "size": 15, "shape": "image-gipuzkoa", "image": gipuzkoa_img64, "color": "#33B5E7-gipuzkoa"},
        {"id": 8, "label": "Gogo", "size": 30, "color": "#FF5733-italy", "shape": "dot", "title": "I am so Alone"},
        {"id": 9, "label": "comecome", "size": 8, "color": "#FF5733-italy", "shape": "dot", "title": "I am so Alone"},
            {"id": 10, "label": "Image Node B", "size": 15, "shape": "image-gipuzkoa", "image": gipuzkoa_img64, "color": "#33B5E7-gipuzkoa"},
            {"id": 11, "label": "Info B", "size": 48, "color": "#33B5E5-canada", "shape": "icon-australia",
            "icon": {"face": "Font Awesome 5 Brands", "code": "\uf05a", "size": 40, "color": "#33B5E5-canada"}},
                {"id": 12, "label": "turku", "size": 15, "shape": "image-turku", "image": turku_img64, "color": "#33B5E6-turku"},
                {"id": 13, "label": "dict", "size": 9, "color": {"background": "yellow-Finland", "border": "green"}, "shape": "dot", "title": "starting the point"},
                {"id":'abcdefg',"label":"abcdefg-label",'shape':'square','size':50},
                
    ]

    edges = [
        {"from": 1, "to": 3, "width": 1, "color": {"color": "#C70039-europe"}, "arrows": "to", "title": "from 1 to 3"},
        {"from": 1, "to": 2, "width": 2, "color": {"color": "#C70038-africa"}},
        {"from": 2, "to": 4, "width": 3, "color": {"color": "#581845-africa"}, "arrows": "to", "dashes": True},
        {"from": 2, "to": 5, "width": 4, "color": {"color": "#FFC300-australia"}, "arrows": "to", "smooth": {"type": "curvedCW"}},
        {"from": 3, "to": 3, "width": 5, "color": {"color": "#DAF7A6-america"}, "arrows": "to"},
        {"from": 4, "to": 6, "width": 6, "color": {"color": "#4CAF50-europe"}, "arrows": "to", "smooth": {"type": "curvedCCW"}},
        {"from": 5, "to": 1, "width": 7, "color": {"color": "#2196F3-asia"}, "arrows": "to", "dashes": True},
        {"from": 7, "to": 1, "width": 8, "color": {"color": "#FF00FF-africa"}, "arrows": "to"},
        {"from":14,"to":13}
    ]
    visnet(nodes, edges, network_title='small Network', browserView=True, writeHTML='test_small.html', min_default_node_size=1, min_default_edge_width=0, maximum_display=10)



    import random 
    nodes2 = []
    edges2 = []

    for i in range(1, 10001):  # Changed to 10,000 nodes
        if i<20:
            size = random.randint(10,20)
        else:
            size = random.randint(1,5)
        width = random.randint(1, 4)
        node = {"id": i, "label": "Node " + str(i), "size": size, "color": random.choice(['red','blue','yellow']), "shape": "dot", "title": "Node " + str(i)}
        nodes2.append(node)
        edge = {"from": random.randint(1, 20), "to": i , "width": width, "color": {"color": "#131112"}, "arrows": "to", "title": f"to {i} width {width}"}
        edges2.append(edge)

    visnet(nodes2, edges2, network_title='Large Network', browserView=True, writeHTML='test_large.html', min_default_node_size=1, min_default_edge_width=1, maximum_display=100)



from spellchecker import SpellChecker
import os,re,spacy
from unidecode import unidecode
import torch
import pandas as pd
from maverick import Maverick
from typing import List
from spacy.language import Language
import unicodedata
from spacy.matcher import PhraseMatcher
from itertools import combinations
from nltk.sentiment import SentimentIntensityAnalyzer
from nltk.corpus import wordnet as wn
import networkx as nx
import ast

# ADJ -ADV, ADJ -NOUN, ADV-VERB, VERB-NOUN 
#text = "girls has extremely wonderful food and drink, boy are loved deeply by grandparents."

# NOUN, PROPN, ENTITY (only get lenght larger than 2 )
# VERB(AUX), ADJ, ADV

# nlp = spacy.load("en_core_web_trf") 

#text = "Harry Potter, the leading tourism scholar, google , Apple company in rovaniemi  has good prodcut in Finland , applies artificial intelligence ,natural language processing to Finnish business. University of Eastern Finland peng(at)hotmail dot com, peng@uef.fi www.uef.fi or uef.fi"
# Define the set of negating lemmas
#NEG_LEMMAS = {"no", "not", "n't", "never", "none", "nothing", "nobody", "no one", "neither","nor"} # nobody, nothing, none will be kept in the results as NOUN 



@Language.component("custom_sentencizer")
def custom_sentencizer(doc):
    """nlp.add_pipe("custom_sentencizer", before="parser")

    doc = nlp("I do not love you, neither you nor me is good, I haven't checked the food yet, how about we are just waiting for mother")

    for sent in doc.sents:
        print("SENT:", sent.text)"""
    for i, token in enumerate(doc):
        # Define words that can start new sentences
        sentence_starters = ["neither", "however", "how", "but", "so",
                             "i", "we", "they", "you", "he", "she", "it","none"]

        # Mark as sentence start if token matches and is not the first token
        if token.text.lower() in sentence_starters and i > 0:
            # Also ensure the previous token is punctuation or comma
            if doc[i - 1].text in [",", ".", "!", "?",";","-","--"]:
                token.is_sent_start = True
    return doc


def spaceOut(text):
    """ make space of emojis"""
    #targets = {',', '.', '!', '?', ';'}
    output = []
    
    for char in text:
        # Check if it's one of your specific punctuation marks
        #if char in targets:
            #output.append(f" {char} ")
        # Check if it's an emoji (Unicode range for most emojis)
        if ord(char) > 0x2000: 
            output.append(f" {char} ")
        else:
            output.append(char)
            
    # Join and clean up double spaces
    return " ".join("".join(output).split())


def word_similarity(word1, word2, pos=wn.NOUN):
    """find two words simliarty 

    Args:
        word1 (string): word1
        word1 (string): word1
        pos (part of speech, optional): wn.VERB,wn.NOUN,wn.ADJ,wn.ADV. Defaults to wn.NOUN.

    Returns:
        foat: word simliarity 0-1 
    """
    synsets1 = wn.synsets(word1, pos=pos)
    synsets2 = wn.synsets(word2, pos=pos)

    # Exit early if either word has no synsets (e.g., not in WordNet or incorrect POS)
    if not synsets1 or not synsets2:
        return 0.0

    max_sim = 0.0
    for s1 in synsets1:
        for s2 in synsets2:
            if s1 is not None:
                sim = s1.wup_similarity(s2)
                
            # This check is crucial: It ignores any sim value that is None
            if sim is not None and sim > max_sim:
                max_sim = sim
    
    return max_sim

# Example usage (assuming 'tree' and 'bush' are Noun synsets)
# print(word_similarity('tree', 'bush', pos=wn.NOUN))




def uniSimilarNodesDF(original_edges, col1 ='node1',col2='node2', pos ='NOUN',minSimilar = 0.91,manual_pairs=None):
    """mapping the high simaliry ndoes in nodes dataframe, e.g. replace time to period , if both in the nodes 

    Args:
        original_edges (_type_): dataframe with columns by node1 and node2 of edge 
        col1,col2 : nodes column names
        pos: simliarty coparison by part of speech, only NOUN, VERB 
        minSimilar: set the minimal value of similarity # food drink 0.909 
        manual_pairs (dict): Optional dictionary of { 'old_value': 'new_value' }
    Return:
           new nodes dataframe replaced simliar words, simlilarity mapping  
    """
    if pos =='NOUN':
        pos1 = wn.NOUN
        pos_cut = '_2noun'
    else:
        raise ValueError(f"Unexpected pos found: '{pos}'. we only take NOUN simliarity, please add other pairs manually")
    noun_nodes = original_edges[col1].tolist() + original_edges[col2].tolist()
    noun_nodes1 = set(noun_nodes)
    noun_nodes2 = [node for node in noun_nodes1 if pos_cut in node]
    

    noun_pairs = list(combinations(noun_nodes2, 2))
 
    #simPairs = {}
    simPairs = manual_pairs.copy() if manual_pairs else {}
    for pair in noun_pairs:
        
        word1 = pair[0][:-6]
        word2 = pair[1][:-6]
        simV = word_similarity(word1,word2, pos=pos1)   # food drink 0.909 
        if simV > minSimilar:
            simPairs.update({pair[0]:pair[1]})

    # Apply the replacement to specific columns
    cols_to_fix = [col1, col2]
    #original_edges[cols_to_fix] = original_edges[cols_to_fix].replace(simPairs)
    original_edges.loc[:, cols_to_fix] = original_edges[cols_to_fix].replace(simPairs)
    simPairs = pd.DataFrame(list(simPairs.items()), columns=['key', 'value'])
    return original_edges, simPairs  





UNICODE_TO_ASCII = {
    '’': "'",   # Right Single Quote (Apostrophe)
    '‘': "'",   # Left Single Quote
    '“': '"',   # Left Double Quote
    '”': '"',   # Right Double Quote
    '–': '-',   # En Dash
    '—': '-',   # Em Dash
    '…': '...', # Ellipsis
    # Add any other non-emoji problematic chars here (e.g., U+00A0 non-breaking space)
}

def convert_specific_unicode(text,UNICODE_TO_ASCII = UNICODE_TO_ASCII):
    """Replaces specific Unicode characters (non-emojis) with their ASCII equivalent."""
    for unicode_char, ascii_char in UNICODE_TO_ASCII.items():
        text = text.replace(unicode_char, ascii_char)
    return text

def merge_emoji_tokens(doc):
    spans = []
    i = 0
    while i < len(doc) - 1:
        # Check if next token is a ZWJ or emoji modifier
        if i + 1 < len(doc) and doc[i+1].text in ['\u200d', '\ufe0f', '♂️', '♀️']:
            start = i
            end = i + 2
            while end < len(doc) and doc[end].text in ['\u200d', '\ufe0f', '♂️', '♀️']:
                end += 1
            
            # If the sequence ended on a ZWJ, grab the next emoji char
            if end < len(doc) and doc[end-1].text == '\u200d':
                end += 1
                
            spans.append(doc[start:end])
            i = end
        else:
            i += 1

    with doc.retokenize() as retokenizer:
        for span in spans:
            # CLEAN THE TEXT: Remove the ZWJ from the merged token's text
            clean_text = span.text.replace('\u200d', '').strip()
            
            # Force the merged token to be a NOUN so semanticEdges likes it
            retokenizer.merge(span, attrs={"ORTH": clean_text, "LEMMA": clean_text, "POS": "NOUN"})
            
    return doc



def is_ascii_word(word):
    try:
        word.encode("ascii")
        return True
    except UnicodeEncodeError:
        return False
spell = SpellChecker()
def spellCorrect(text,nonCheckWords = ['offf']):
    """_summary_

    Args:
        text (_type_): _description_
        nonCheckWords (list, optional): _description_. Defaults to ['offf'].

    Returns:
        _type_: _description_
    """
    
    
    # Split text into words and punctuation
    tokens = text.split()
    spell.word_frequency.load_words(nonCheckWords)
    corrected_tokens = []
    for token in tokens:
        if token.isalpha() and is_ascii_word(token):
            corrected = spell.correction(token)
            
            corrected_tokens.append(corrected if corrected else token)
        elif len(token)>0:
            corrected_tokens.append(token)
            
        
    return " ".join(
        corrected_tokens
                    )



class MaverickCoref:
    """
    word coreference 
    # Make sure dependencies are installed:
    # pip install torch transformers hydra-core pytorch-lightning
    # pip install --no-deps git+https://github.com/Atsaniik/maverick-coref.git

    # hf download sapienzanlp/maverick-mes-preco --local-dir ./maverick-mes-preco
    # hf download sapienzanlp/maverick-mes-ontonotes --local-dir ./maverick-ontonotes

    # import nltk
    # nltk.download('punkt_tab')

    # To uninstall existing torch (if using pip)
    # pip uninstall torch torchvision torchaudio
    # pip install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu124 , instll the cuda 

    """
    _model = None
    _device = None
    
    # Store the path used to load the current cached model
    _loaded_path = None 

    @classmethod
    def _ensure_model_loaded(cls, path=None, device=None):
        """
        Internal method to load and cache the model ONLY if it hasn't been loaded yet 
        OR if a different path/device is requested.
        """
        # Define the path to use for this check
        default_path = os.path.join(os.getcwd(), "sapienzanlp", "maverick-ontonotes")
        path_to_use = path if path is not None else default_path
        
        # 1. Auto-detect CUDA only if device is not explicitly set
        if device is None:
            device_to_use = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            device_to_use = device

        # 2. Load model ONLY if it's the first time OR if the device/path has changed
        if cls._model is None or cls._device != device_to_use or cls._loaded_path != path_to_use:
            cls._device = device_to_use
            cls._loaded_path = path_to_use # Update the stored path
            print(f"Loading Maverick model from {path_to_use} on {cls._device.upper()}...")
            
            try:
                cls._model = Maverick(
                    hf_name_or_path=path_to_use,
                    device=cls._device
                )
            except Exception as e:
                print(f"Error loading model from path: {path_to_use}. Error: {e}")
                # Reset cache variables if loading fails
                cls._model = None
                cls._device = None
                cls._loaded_path = None
                raise # Re-raise the exception to stop execution

    @classmethod
    def resolve(cls, text, device=None, path=None)-> str: # Added 'path' argument
        """
        Resolves coreference chains in the input text using the cached model.

        Args:
            text (str): The input text string.
            device (str, optional): The device to use. Defaults to auto-detect.
            path (str, optional): The path to the Maverick model. If provided, 
                                  it will override the default path and force a reload 
                                  if the model is not already loaded from this location.

        Returns:
            str: The text with coreferences resolved.
        """
        # 1. Ensure the model is loaded, passing both device and the optional path
        # This will trigger a reload if the model in cache was loaded from a different path.
        cls._ensure_model_loaded(path=path, device=device) 
        
        # 2. Directly access the cached model, which is guaranteed to be loaded here
        model = cls._model
        assert model is not None
        # 3. Proceed with prediction using the model instance
        result = model.predict(text)
        resolved_tokens = result['tokens'][:]

        for cluster in result['clusters_token_offsets']:
            first_mention = cluster[0]
            first_mention_text = " ".join(result['tokens'][first_mention[0]:first_mention[1]+1])
            
            for mention in cluster[1:]:
                start, end = mention
                resolved_tokens[start] = first_mention_text 
                for i in range(start+1, end+1): 
                    resolved_tokens[i] = ""

        return " ".join([tok for tok in resolved_tokens if tok])
   

def replace_word(text, old, new, case_insensitive=True):
    "replace strictly of words of old to new but protect the other forms such as old.com , older"
    flags = re.IGNORECASE if case_insensitive else 0
    pattern = rf'(?<![\w@./?&=#-]){re.escape(old)}(?![\w.])'
    return re.sub(pattern, new, text, flags=flags)



def verbPhrase(doc):
    """
    Detects compact phrasal and prepositional verbs like:
    'give up', 'take off', 'look after', 'put up with', 'give away', 'turn off'.
    Returns list of base-form verb phrases (lemma-based) and new text 
    """
    
    #doc = nlp(text)
    new_text =  doc.text
    results = []
    valid_particles = {
        "up", "off", "away", "out", "over", "back",
        "down", "on", "in", "with", "into", "by","after"
    }

    for token in doc:
        if token.pos_ != "VERB":
            continue

        phrase_parts = [token.lemma_]
        found = False

        # --- Direct particle or preposition ---
        for child in token.children:
            if (
                child.text.lower() in valid_particles
                and child.dep_ == "prt" #   child.dep_ == in {"prt", "prep"} deprecated on 6.01.2026
            ):
                phrase_parts.append(child.text)
                #print('new_phrase', child.text,doc[token.i +1])
                if child.text != doc[token.i +1].text:
                    
                    new_text = new_text.replace(doc[child.i].text,"").replace(doc[token.i].text, " ".join([token.lemma_,child.text]))
                    
                found = True

        if found:
            results.append(" ".join(phrase_parts))
            continue



    return  list(dict.fromkeys(results)), new_text




def is_emoji(token):
    # Emojis are symbols with Unicode category starting with 'So'
    return any(unicodedata.category(char).startswith('So') for char in token.text)

def cleanEEL(doc,cleanEMOJI = False):
    """ remove the emojis, emails, urls 
    Args:
        doc (doc): spacy doc 
        nlp (lanaguge): spacy nlp model , for better detection esepaiclly emails and urls , spacy.load("en_core_web_trf") is recommended

    Returns:
        tuple[str, list[str], list[str], list[str]]:
            (cleaned_text, emojis, emails, urls)
    """
    #doc = nlp_trf(text)
    cleaned_tokens = []
    emails = []
    urls = []
    emojis = []
    for token in doc:
        # Skip if it's an email, URL, or emoji
        if token.like_email:
            emails.append(token.text)
            continue
        if token.like_url:
            urls.append(token.text)
            continue
        if cleanEMOJI:
            if is_emoji(token):
                emojis.append(token.text)
                continue

        cleaned_tokens.append(token.text)

    return " ".join(cleaned_tokens), emojis, emails, urls 

    # Example
    #sample_text = "Contact me at hello@sample.com 😄 or visit https://example.com for details! Peng Yang, the leading tourism scholar, applies natural language processing to  Finnish business. peng(at)hotmail dot com, peng@uef.fi www.uef.fi or uef.fi"
    #cleaned,emojis, emails, urls = cleanEEL(sample_text,nlp)
    #print(cleaned,emojis, emails, urls,sep="\n" )


def cleanDetNum(text,nlp:Language, det=["a", "an", "the"],numberWords =True ):
    """remove the Determiner a , an ,the and numbers but keep COVID-19, G20

    Args:
        text (string): any text 
        det (list, optional):det. Defaults to ["a", "an", "the"].
        numberWords(bool): if True, will remove any numbers including words such as one, two , zero. 

    Returns:
        str: new text without determiner or numbers
    """
    doc = nlp(text)
    cleaned_tokens = []

    for token in doc:
        lower = token.text.lower()

       
        if token.pos_ == "DET" and lower in det:
            continue
        
        if numberWords:
            if (
                token.like_num or
                re.fullmatch(r"\d+", token.text)
            ):
                # But keep tokens that have letters or hyphens (like COVID-19, G20)
                if not re.search(r"[A-Za-z-]", token.text):
                    continue
                if re.match(r".*[A-Za-z]+-\d+.*", token.text):
                    pass  # keep COVID-19
                else:
                    continue

        cleaned_tokens.append(token.text)

    return " ".join(cleaned_tokens)

    # Example
    #text = "The COVID-19 pandemic G200 20 kids , 3.14 one trillion lasted thirty one months and affected 9 million people in a year."
    #print(cleanDetNum(text,nlp2,numberWords=False))



def cleanText(text, nlp, det=None, valid_particles =None, number_words=True, clean_emoji=False, heavy_number = False,verbPhrase = False, extraClean = False
              ,reADJ=['meilide'], reADV=['piaoliangdi'],reVERB=['caiquxingdong'], 
              reNOUN=['Natural Language Processing'],reENT = True,hyphenated = True,):
    """
    Integrates text cleaning (removing determiners, numbers, URLs, emails) 
    with phrasal verb detection and normalization.
    
    Args:
        text (str): The raw input text.
        nlp (spacy.lang): Loaded Spacy language model.
        det (list): List of determiners to remove.
        verbPhrase: if true, then 'take it off' beacome 'take off it' in the new text
        valid_particles(set): valid_particles = {
                            "up", "off", "away", "out", "over", "back",
                            "down", "on", "in", "with", "into", "by", "after"
                                                                            }
        heavy_number(bool): If True, remove all numbers including COVID-19, G20, etc
        number_words (bool): If True, remove numbers (keeping COVID-19, G20, etc).
        clean_emoji (bool): If True, removes emojis.
        
    Returns:
        tuple: (cleaned_text, verb_phrases_list, emails_list, urls_list, emojis_list)
    """
    if det is None:
        det = ["a", "an", "the"]
    text = re.sub(r'([()\[\]{}])', r' \1 ', text)  # make space before and after both (,)
    #doc = nlp(text)
    doc = docRetoken(text, nlp,reADJ=reADJ, reADV=reADV,
                         reVERB=reVERB , reNOUN=reNOUN,
                         reENT = reENT,hyphenated = hyphenated)
    
    # --- Output Containers ---
    cleaned_tokens = []
    verb_phrases_found = []
    emails = []
    urls = []
    emojis = []
    
    # --- Phrasal Verb Configuration ---
    if valid_particles is None:
        valid_particles = {
            "up", "off", "away", "out", "over", "back",
            "down", "on", "in", "with", "into", "by", "after"
        }
    
    
    # Maps a verb's token index to its new string (e.g., "take off")
    verb_replacements = {} 
    # Set of token indices (particles) to skip during reconstruction
    indices_to_skip = set() 

    # ---------------------------------------------------------
    # PASS 1: Identify Phrasal Verbs (Dependency Parsing)
    # ---------------------------------------------------------
    if verbPhrase:
        for token in doc:
            if token.pos_ == "VERB":
                # Check children for particles
                for child in token.children:
                    if (child.text.lower() in valid_particles 
                        and child.dep_ == "prt"):   #  child.dep_ == "prt" updated 6.jan.2026 , child.dep_ in {"prt", "prep"}) 
                        
                        # We found a phrasal verb (e.g., "took ... off")
                        lemma_phrase = f"{token.lemma_} {child.text.lower()}"
                        
                        # Store the result
                        verb_phrases_found.append(lemma_phrase)
                        
                        # Logic for text reconstruction:
                        # 1. Mark the particle to be skipped later
                        indices_to_skip.add(child.i)
                        
                        # 2. Mark the verb to be replaced by "lemma + particle"
                        # Note: This unifies split verbs. "took it off" -> "take off it"
                        verb_replacements[token.i] = lemma_phrase
                        
                        # Only handle one particle per verb to avoid conflicts
                        break 

    # ---------------------------------------------------------
    # PASS 2: Filter Tokens and Reconstruct Text
    # ---------------------------------------------------------
    for token in doc:
        # 1. specific logic to skip particles of phrasal verbs
        if token.i in indices_to_skip:
            continue
            
        # 2. Extract specific entities
        if token.like_email:
            emails.append(token.text)
            #print(f'email:{token.text}')
            continue # Remove email from text? Remove this 'continue' if you want to keep it.
            
        if token.like_url:
            urls.append(token.text)
            #print(f'url:{token.text}')
            continue # Remove URL from text? Remove this 'continue' if you want to keep it.
            
        # 3. Emoji Logic
       
        # Simple check: if non-ascii and character category is Other Symbol (So)
        if is_emoji(token): 
            emojis.append(token.text)
            #print(f'emoji: {token.text}')
            if clean_emoji:
                    continue

        # 4. Determiner Logic
        if token.pos_ == "DET" and token.text.lower() in det:
            continue
            
        # 5. Number Logic
        if heavy_number:
            if number_words:
                is_number = (token.like_num or re.fullmatch(r"\d+", token.text))
                if is_number:
                    # Preservation Logic for G20, COVID-19
                    has_letters = re.search(r"[A-Za-z]", token.text)
                    is_compound = re.match(r".*[A-Za-z]+-\d+.*", token.text) # e.g. COVID-19
                    
                    if not has_letters: 
                        continue # Pure number -> skip
                    elif not is_compound and not has_letters:
                        continue # Redundant check, but keeps logic clear
                    
      
        else:
            if number_words:
                if token.like_num:
                    continue

                # contains any digits?
                if any(ch.isdigit() for ch in token.text):
                    # keep if contains letters OR hyphens → "COVID-19", "G20"
                    if any(ch.isalpha() for ch in token.text) or "-" in token.text:
                        pass  # keep
                    else:
                        continue
      
      
        #print('clean_text process --------', token.text )
        # 6. Text Reconstruction
        if verbPhrase:
            if token.i in verb_replacements:
                # This is the verb part of a phrasal verb -> Insert the full lemma phrase
                cleaned_tokens.append(verb_replacements[token.i])
            else:
                # Standard token -> Insert text
                cleaned_tokens.append(token.text)
        else:
            # Standard token -> Insert text
            cleaned_tokens.append(token.text)

    # Remove duplicates from results while preserving order
    unique_phrases = list(dict.fromkeys(verb_phrases_found))
    
    # Join cleaned text
    final_text = " ".join(cleaned_tokens)
    
    
    if extraClean:
        #final_text = re.sub(r'\s+([?.!,"])', r'\1', final_text) # Cleanup extra spaces around punctuation (optional polish)
        final_text = re.sub(r'\(.*?\)', '', final_text)  # remove () Remove parentheses and everything inside them
        final_text = re.sub(r'\s+', ' ', final_text).strip()

    return final_text, unique_phrases, emails, urls, emojis







def docRetoken(text, nlp, reADJ=['meilide'], reADV=['piaoliang'],reVERB=['caiquxingdong'], reNOUN=['natural language processing'],reENT = True,hyphenated = True):
    """retokenize specific words: adj,adv,verb, noun and entity 

    Args:
        text (str): _description_
        nlp (spacy nlp): _description_
        reNOUN (list, optional): list of noun to be retonized. Defaults to ['natural language processing'].
        reADJ (list, optional): _list of adj to be retonized. Defaults to ['meilide'].
        reADV (list, optional): list of adv to be retonized. Defaults to ['piaoliang'].
        reVERB (list, optional): list of verb to be retonized. Defaults to ['caiquxingdong'].
        reENT (bool, optional): if also retonize entity to noun. Defaults to True.
        hypenated(bool,opetional): if true, tokenie the hypenated word but POS it depends on the root 

    Returns:
        doc: new doc after retokenziation 
    """
    
    pinyin_rules = {
    "ADV": [nlp.make_doc(x) for x in reADV ],
    "VERB": [nlp.make_doc(x) for x in reVERB ],
    "ADJ": [nlp.make_doc(x) for x in reADJ ],
    "NOUN":[nlp.make_doc(x) for x in reNOUN]
     }
    

    # Create the PhraseMatcher
    matcher = PhraseMatcher(nlp.vocab)
    for pos_tag, patterns in pinyin_rules.items():
        matcher.add(pos_tag, patterns)

    # --- 2. Process the text ---
    doc = nlp(text)

    # Find all pinyin matches
    pinyin_matches = matcher(doc)

    # --- 3. Prepare all spans for merging ---
    # We will collect all spans (pinyin + entities) to merge them in one go.
    # We store them as (span, pos_tag)
    spans_to_merge = []
    
    if hyphenated:
        hyphenated_matches = re.finditer(r'\b\w+(?:-\w+)+\b', text)

        for match in hyphenated_matches:
            span = doc.char_span(match.start(), match.end())
            if span:
                #print(f"Span: {span.text}, POS: {span.root.pos_}")
                spans_to_merge.append((span, span.root.pos_))

    # Add pinyin matches
    for match_id, start, end in pinyin_matches:
        pos_tag = nlp.vocab.strings[match_id]
        span = doc[start:end]
        spans_to_merge.append((span, pos_tag))

    # Add entity matches (Rule: "turn all entity to noun")
    if reENT:
        for ent in doc.ents:
            # 'Peng Yang' will be found as an entity (PERSON)
            # We add it to our list with the target POS "NOUN"
            spans_to_merge.append((ent, "NOUN"))

        # Sort spans by their start index. This is crucial for retokenizing correctly.
        spans_to_merge.sort(key=lambda x: x[0].start)

    # --- 4. Retokenize in one step ---
    # Use doc.retokenize() to merge all collected spans
    with doc.retokenize() as retokenizer:
        for span, pos_tag in spans_to_merge:
            try:
                # Set the POS and TAG attributes for the merged token
                retokenizer.merge(span, attrs={"POS": pos_tag, "TAG": pos_tag})
            except ValueError:
                # This can happen if spans overlap (they don't in this case)
                # We can just skip the conflicting merge
                pass
    return doc



def get_all_conjuncts(token):
    """
    Helper function to return a list of a token and all its conjuncts.
    """
    return [token] + list(token.conjuncts)








def semanticEdges(doc,NEG_Lemmas = {"no", "not", "n't", "never", "none", "nothing", "nobody", "neither","nor"},
                            nonSemantic = 0, NN = 'notInclude',copulaBridge = True):
    """
    Analyzes doc and extracts a set of ANNOTATED semantic edges
    with POS/NEG negation labels.
    
    Note: verb and noun should put all forms and casesensetive , such as turn off, turned off, turn off; countries, country , natural language processing, Natural Language Processing
    Note2: the edge with nothing, nobody also show NEG, nothing-good-NEG then no need add label extra NEG. 
    Args:
        doc (doc): spacy nlp tokenzied doc 
        
        NEG_lemaas(set,optional): set of negating lemaas. Defaults to {"no", "not", "n't", "never", "none", "nothing", "nobody", "no one", "neither","nor"}
        nonSemantic(digit,optional): if nonSemantic = 1, then generate avn edges of non semantic text adj-adv, adj-noun, adv-verb, verb-noun 
        NN: (vaule: [inlcude, only , notInclude] ) noun-noun type nonSemantic edges, othervise A-A-V-N edge
        copulaBridge (bool): Finland is beautiful add one extra edge Finland-beautiful connected by copula Be , otherwise only Finland-be, be-beautiful 
    Returns:
        list: semantic edges with Negation labels  
    """

    
    
    #doc = docRetoken(text,nlp,reADJ=reADJs, reADV=reADVs,reVERB=reVERBs, reNOUN=reNOUNs,reENT = reENTs,hyphenated=reHyphens)
    
    edges = set()
    if 'none' in doc.text.lower():
        suffix ="NEG"
    else:
        suffix = "POS"
    for token in doc:
        dep = token.dep_
        head = token.head
      
        # Rule 1: NOUN-VERB (Subjects)
        # e.g., "girls have", "boy is loved", "nobody likes"
        if dep in ('nsubj', 'nsubjpass') and head.pos_ in ('VERB', 'AUX','SYM'):  # add 'SYM' 18.12.2025
            subjects = get_all_conjuncts(token)
            verbs = get_all_conjuncts(head)
            for s in subjects:
                for v in verbs:
                    if v.i > 0 and v.nbor(-1).lemma_.lower() in NEG_Lemmas:
                        suffix = "NEG"
                    elif s.i > 0 and s.nbor(-1).lemma_.lower() in NEG_Lemmas:  # neither you nor me like the food
                        suffix = "NEG"
                    edges.add(f"{s.lemma_}-{v.lemma_} (NOUN-VERB {suffix})")

        # Rule 2: VERB-NOUN (Objects)
        # e.g., "have food", "don't eat avocado", "none like coffee"
        #elif dep == 'dobj' and head.pos_ == 'VERB':
        elif dep in ('dobj', 'pobj', 'attr', 'dative') and head.pos_ in ('VERB', 'AUX','SYM'): # add 'SYM' 18.12.2025
            objects = get_all_conjuncts(token)
            verbs = get_all_conjuncts(head)
            
            # Find all subjects of the verb
            subjects = [c for c in head.children if c.dep_ in ('nsubj', 'nsubjpass')]
            
            for o in objects:
                for v in verbs:
                    if v.i > 0 and v.nbor(-1).lemma_.lower() in NEG_Lemmas:
                        suffix = "NEG"
                    edges.add(f"{v.lemma_}-{o.lemma_} (VERB-NOUN {suffix})")
                    
            # Rule 7: COMPREHENSIVE COPULA BRIDGE (Subject -> Attribute's Adjective)
            # Handles: "We become beautiful girls", "Finland is a safe country" to have We-beautiful , Finland-safe
            if dep == 'attr' and head.pos_ in ('VERB', 'AUX') and copulaBridge:
                # 1. Find the subjects of the copular verb (e.g., "We", "Finland")
                subjects = [c for c in head.children if c.dep_ in ('nsubj', 'nsubjpass')]
                
                # 2. Find adjectives modifying this attribute noun (e.g., "beautiful", "safe")
                # We look for 'amod' (adjectival modifiers)
                adjectives = [c for c in token.children if c.dep_ == 'amod']
                
                for s in subjects:
                    subj_conj = get_all_conjuncts(s)
                    for sc in subj_conj:
                        for adj in adjectives:
                            adj_conj = get_all_conjuncts(adj)
                            for ac in adj_conj:
                                local_suffix = suffix
                                # Check for local negation (e.g., "We become not-so-beautiful girls")
                                if ac.i > 0 and ac.nbor(-1).lemma_.lower() in NEG_Lemmas:
                                    local_suffix = "NEG"
                                
                                # Adding as NOUN-ADJ per your request (even if the subject is a pronoun like 'we')
                                edges.add(f"{sc.lemma_}-{ac.lemma_} (NOUN-ADJ {local_suffix})")                
    
    
    

        # Rule 3: ADJ-NOUN (Adjectival Modifier)
        # e.g., "wonderful food", "no tasty food"
        elif dep == 'amod' and head.pos_ == 'NOUN':
            adjectives = get_all_conjuncts(token)
            nouns = get_all_conjuncts(head)
            for a in adjectives:
                
                for n in nouns:
                    if a.i > 0 and a.nbor(-1).lemma_.lower() in NEG_Lemmas:
                        suffix = "NEG"
                        
                    edges.add(f"{a.lemma_}-{n.lemma_} (ADJ-NOUN {suffix})")

        # Rule 4: ADV-VERB / ADV-ADJ (Adverbial Modifier)
        elif dep == 'advmod':
            adverbs = get_all_conjuncts(token)
            
            # e.g., "loved deeply", "not really good"
            if head.pos_ in ('VERB', 'AUX'):
                verbs = get_all_conjuncts(head)
                for adv in adverbs:
                    for v in verbs:
                        if v.i > 0 and v.nbor(-1).lemma_.lower() in NEG_Lemmas:
                            suffix = "NEG"
                    edges.add(f"{adv.lemma_}-{v.lemma_} (ADV-VERB {suffix})")
                
            # e.g., "extremely wonderful", "not very good"
            elif head.pos_ == 'ADJ':
                adjectives = get_all_conjuncts(head)
                for adv in adverbs:
                    for adj in adjectives:
                        if adj.i > 0 and adj.nbor(-1).lemma_.lower() in NEG_Lemmas:
                            suffix = "NEG"
        
                        edges.add(f"{adv.lemma_}-{adj.lemma_} (ADV-ADJ {suffix})")

        # Rule 5: VERB-NOUN (Prepositional Objects & Agents)
        # e.g., "go to school", "is not in house"
        elif dep == 'pobj' and head.pos_ == 'ADP':
            verb = head.head
            if verb.pos_ in ('VERB', 'AUX') and head.dep_ in ('prep', 'agent'):
                objects = get_all_conjuncts(token)
                verbs = get_all_conjuncts(verb)
                
                # Find all subjects of the verb
                subjects = [c for c in verb.children if c.dep_ in ('nsubj', 'nsubjpass')]

                for v in verbs:
                    for o in objects:
   
                        if v.i > 0 and v.nbor(-1).lemma_.lower() in NEG_Lemmas:
                            suffix = "NEG"
                        edges.add(f"{v.lemma_}-{o.lemma_} (VERB-NOUN {suffix})")

        # Rule 6: NOUN-ADJ (Predicative Adjectives)
        # e.g., "food is wonderful", "food is not good"
        elif dep == 'acomp' and head.pos_ in ('VERB', 'AUX'):
            adjectives = get_all_conjuncts(token)
            subjects = [child for child in head.children if child.dep_ in ('nsubj', 'nsubjpass')]
            for s in subjects:
                subject_conj = get_all_conjuncts(s)
                for sc in subject_conj: 
                    for a in adjectives:
                        if a.i > 0 and a.nbor(-1).lemma_.lower() in NEG_Lemmas:
                            suffix = "NEG"
                        edges.add(f"{sc.lemma_}-{a.lemma_} (NOUN-ADJ {suffix})")
        
    if NN in ['only','include']:
        nonSemantic = 1                 
    # --- Add default edges if none found ---
    if  nonSemantic ==1:
        # Count negation lemmas
        edgesSemantic = [edge.split("(")[0].strip() for edge in edges]
        neg_count = sum(1 for token in doc if token.lemma_ in NEG_Lemmas)
        is_neg = (neg_count % 2 == 1)
        suffix = "NEG" if is_neg else "POS"

        # Filter out negation tokens
        valid_tokens = [t for t in doc if t.lemma_ not in NEG_Lemmas]

        advs = [t.lemma_ for t in valid_tokens if t.pos_ == "ADV"]
        adjs = [t.lemma_ for t in valid_tokens if t.pos_ == "ADJ"]
        verbs = [t.lemma_ for t in valid_tokens if t.pos_ == "VERB"]
        nouns = [t.lemma_ for t in valid_tokens if t.pos_ == "NOUN"]


        if NN=='only':
            # Get all combinations of size 2
            print('---- only NN non semantic edges added ----')
            all_pairs = list(combinations(nouns, 2))
            for nn in all_pairs:
                if f"{nn[0]}-{nn[1]}" not in edgesSemantic:
                    edges.add(f"{nn[0]}-{nn[1]} (NOUN-NOUN {suffix} NonSemantic)")
        elif NN =='notInclude':
            print('---- no NN but AAVN non semantic edges added ----')
            # ADV-ADJ
            for adv in advs:
                for adj in adjs:
                    if f"{adv}-{adj}" not in edgesSemantic:
                        edges.add(f"{adv}-{adj} (ADV-ADJ {suffix} NonSemantic)")

            # ADV-VERB
            for adv in advs:
                for verb in verbs:
                    if f"{adv}-{verb}" not in edgesSemantic:
                        edges.add(f"{adv}-{verb} (ADV-VERB {suffix} NonSemantic)")

            # ADJ-NOUN
            for adj in adjs:
                for noun in nouns:
                    if f"{adj}-{noun}" not in edgesSemantic:
                        edges.add(f"{adj}-{noun} (ADJ-NOUN {suffix} NonSemantic)")

            # VERB-NOUN
            for verb in verbs:
                for noun in nouns:
                    if f"{verb}-{noun}" not in edgesSemantic:
                        edges.add(f"{verb}-{noun} (VERB-NOUN {suffix} NonSemantic)")
        elif NN =='include':
            print('---- all NN and AAVN non semantic edges added ----')
                        # ADV-ADJ
            for adv in advs:
                for adj in adjs:
                    if f"{adv}-{adj}" not in edgesSemantic:
                         edges.add(f"{adv}-{adj} (ADV-ADJ {suffix} NonSemantic)")

            # ADV-VERB
            for adv in advs:
                for verb in verbs:
                    if f"{adv}-{verb}" not in edgesSemantic:
                        edges.add(f"{adv}-{verb} (ADV-VERB {suffix} NonSemantic)")

            # ADJ-NOUN
            for adj in adjs:
                for noun in nouns:
                    if f"{adj}-{noun}" not in edgesSemantic:
                        edges.add(f"{adj}-{noun} (ADJ-NOUN {suffix} NonSemantic)")

            # VERB-NOUN
            for verb in verbs:
                for noun in nouns:
                    if f"{verb}-{noun}" not in edgesSemantic:
                        edges.add(f"{verb}-{noun} (VERB-NOUN {suffix} NonSemantic)")
                    
            all_pairs = list(combinations(nouns, 2))
            for nn in all_pairs:
                if f"{nn[0]}-{nn[1]}" not in edgesSemantic:
                    edges.add(f"{nn[0]}-{nn[1]} (NOUN-NOUN {suffix} NonSemantic)")

    return edges



def renameAAVN(node1,pospe1):
    """some words can be both noun,verb or adj and adv, need to specify them such as LOVE, light 

    Args:
        node1 (_type_): node lable
        pospe1 (_type_): node part of speech

    Returns:
        _type_: _description_
    """
    node1 = node1.strip()
    if pospe1 == 'ADJ':
        node1b = node1+'_2adj'
    elif pospe1 == 'ADV':
        node1b = node1 +'_2adv'
    elif pospe1 == 'VERB':
        node1b = node1 +'_2verb'
    elif pospe1 == 'NOUN':
        #print("renameNoun", node1)
        node1b = node1 +'_2noun'
    return node1b
        
renameAAVN('love','VERB')      



sia = SentimentIntensityAnalyzer()
def edges2DF(edges,sentiScore = True):
    node1s_rn = []
    node2s_rn = []
    node1s = []
    node2s = []
    weights = []
    pospe1s = []
    pospe2s = []
    semanORnots = []
    sentis = []
    
    for edge in edges:
        #print('edge',edge)
        #text_to_split = "karelian-nothing (ADJ-NOUN POS)"
        split_pattern = r"\(|\)"
        split_result = re.split(split_pattern, edge)
        #print(split_result)
        
        nodes = split_result[0].strip().rsplit('-', 1)   # "Onze-Lieve-Vrouwekerk-be" to parts[0] -> "Onze-Lieve-Vrouwekerk"  parts[1] -> "be"
        #print(nodes)
        node1 = nodes[0]
        node2 = nodes[1]
        pospos = split_result[1].strip().split(" ")
        pospes = pospos[0].strip().split("-")
        #print(pospes)
        pospe1 = pospes[0]
        pospe2 = pospes[1]
        #print(pospes)
        weight0 = pospos[1]
        if weight0 =='POS':
            weight = 1
        elif 'nothing' in nodes or 'nobody' in nodes:
            weight = 1
        else:
            weight = -1
        weights.append(weight)
        if sentiScore:
            senti = sia.polarity_scores(f"{node1} {node2}")['compound']
            sentis.append(senti*weight)
        
        
        if len(pospos)>2:
        
            semanORnot = 'non_semantic'
        else:
            semanORnot = 'semantic'

        node1_rn = renameAAVN(node1,pospe1)
        node2_rn = renameAAVN(node2,pospe2)
        node1s_rn.append(node1_rn)
        node2s_rn.append(node2_rn)
            
        
        #print(node1,node2,pospe1, pospe2, semanORnot)

        node1s.append(node1)
        node2s.append(node2)
        pospe1s.append(pospe1)
        pospe2s.append(pospe2)
        semanORnots.append(semanORnot)
    if sentiScore:
        edgesDF = pd.DataFrame({'node1':node1s_rn,
                        'node2':node2s_rn,
                        'node1a':node1s,
                        'node2a':node2s,
                        
                        'weight':weights,
                        'pospe1':pospe1s,
                        'pospe2':pospe2s,
                        'semanORnot':semanORnots,
                        'sentiment':sentis})
    else:   
        edgesDF = pd.DataFrame({'node1':node1s_rn,
                                'node2':node2s_rn,
                                'node1a':node1s,
                                'node2a':node2s,
                                'weight':weights,
                                'pospe1':pospe1s,
                                'pospe2':pospe2s,
                                'semanORnot':semanORnots})
    return edgesDF
        





#weighted_edges = topicNOUN.noun_edges.tolist()

def preVisnet(weighted_edges, colorNOUN = 'skyblue', shapeNOUN= 'box',
             colorADJ = 'gold', shapeADJ = 'triangle',
             colorVERB = 'pink', shapeVERB = 'square',
             colorADV ='brown', shapeADV = 'star',
             zeroWeightC = 'black',posWeightC ='green',negWeightC ='red',):
    """add sentiment and prepare the visual of aavn weighted edges of topic dataframe e.g. topicNOUN, topicCOMM
    Args:
        weighted_edges (_type_): aavn edges 
        colorNOUN (str, optional): _description_. Defaults to 'skyblue'.
        shapeNOUN (str, optional): _description_. Defaults to 'box'.
        colorADJ (str, optional): _description_. Defaults to 'gold'.
        shapeADJ (str, optional): _description_. Defaults to 'triangle'.
        colorVERB (str, optional): _description_. Defaults to 'pink'.
        shapeVERB (str, optional): _description_. Defaults to 'square'.
        colorADV (str, optional): _description_. Defaults to 'brown'.
        shapeADV (str, optional): _description_. Defaults to 'star'.
        zeroWeightC (str, optional): _description_. Defaults to 'black'.
        posWeightC (str, optional): _description_. Defaults to 'green'.
        negWeightC (str, optional): _description_. Defaults to 'red'.

    Returns:
        _type_: _description_
    """
    
    G = nx.Graph() 
    
    if isinstance(weighted_edges,str):
        weighted_edges1 = ast.literal_eval(weighted_edges)
    else:
        weighted_edges1 = weighted_edges
        
    

    G.add_weighted_edges_from(weighted_edges1)
  
        

    # Node Degrees
    nodeDegree = dict(G.degree(weight=True)) # type: ignore
    nodeDegreeDF = pd.DataFrame({'node': list(nodeDegree.keys()), 'nodeDegree': list(nodeDegree.values())}) 
    nodeDegreeDF = nodeDegreeDF.sort_values(by='nodeDegree', ascending=False)



    # --- 6. Visualization Data ---
    visNodes = []
    for node, degree in zip(nodeDegreeDF.iloc[:, 0], nodeDegreeDF.iloc[:, 1]):
        
        if "_2noun" in node:
            
            color = colorNOUN; shape= shapeNOUN; title= node
        elif "_2adj" in node:
            color = colorADJ; shape = shapeADJ; title = node
        elif "_2verb" in node:
            color = colorVERB; shape = shapeVERB; title = node
        elif "_2adv" in node:
            color =colorADV; shape = shapeADV; title = node 
        else:
            color = 'black'; shape = 'dot'; title = 'non AAVN'
        
        visNode = {"id": node, "label": node.split("_2")[0], "size": degree, "color": color, "shape": shape, "title": f"{title}-{degree}"}
        visNodes.append(visNode)
        
    visEdges = []

    # Calculate global sentiment for coloring
    #global_sentiment = edgesDF.groupby(['node1', 'node2']).agg(sentiment=('sentiment', lambda x: x.mode().iat[0])).reset_index()
    #edges_viz = pd.merge(edgesDF_global, global_sentiment, on=['node1','node2'])

    for edge in  weighted_edges1:
      
        node1  = edge[0]
        node2 = edge[1]
        weight = edge[2]
        #print(node1,node2,weight)
        senti = sia.polarity_scores(f"{node1.split("_2")[0]} {node2.split("_2")[0]}")['compound']
        #print('edge, senti',edge, senti)
        if weight<0:
            sentiment = senti*-1
        else:
            sentiment = senti 
        if weight == 0:
            color = zeroWeightC
        elif weight > 0:
            color =posWeightC
        else:
            color = negWeightC
            
        visEdge = {"from": node1, "to": node2, "width": abs(weight), "color": {"color": color}, "arrows": "to", "title": f"weight-sentiment: {weight}-{sentiment}"}
        visEdges.append(visEdge)
        
    return visNodes, visEdges,nodeDegreeDF 




if __name__ == "__main__":
    nlp = spacy.load("en_core_web_sm") 
    text = "Kuopio is known for its kalakukko (fish pie)"
    edges = semanticEdges(nlp(text))
    print(f"text: {text}")
    print('---edges---')
    for edge in sorted(edges):
        
        print(edge)
    
